# Generated by Django 3.2.13 on 2022-07-22 16:31

from django.db import migrations, models


def data_bytes_to_string_up(apps, schema_editor):
    CollectionVersionSignature = apps.get_model("ansible", "CollectionVersionSignature")
    signature_batch = []
    for signature in CollectionVersionSignature.objects.all():
        signature.new_data = bytes(signature.data).decode()
        signature_batch.append(signature)
        if len(signature_batch) >= 1000:
            CollectionVersionSignature.objects.bulk_update(signature_batch, ["new_data"])
            signature_batch.clear()
    CollectionVersionSignature.objects.bulk_update(signature_batch, ["new_data"])


def data_bytes_to_string_down(apps, schema_editor):
    CollectionVersionSignature = apps.get_model("ansible", "CollectionVersionSignature")
    signature_batch = []
    for signature in CollectionVersionSignature.objects.all():
        signature.data = signature.new_data.encode()
        signature_batch.append(signature)
        if len(signature_batch) >= 1000:
            CollectionVersionSignature.objects.bulk_update(signature_batch, ["data"])
            signature_batch.clear()
    CollectionVersionSignature.objects.bulk_update(signature_batch, ["data"])


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0042_ansiblerepository_gpgkey'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collectionversionsignature',
            name='data',
            field=models.BinaryField(null=True),
        ),
        migrations.AddField(
            model_name='collectionversionsignature',
            name='new_data',
            field=models.TextField(default=''),
            preserve_default=False,
        ),
        migrations.RunPython(
            code=data_bytes_to_string_up,
            reverse_code=data_bytes_to_string_down,
        ),
        migrations.RemoveField(
            model_name='collectionversionsignature',
            name='data',
        ),
        migrations.RenameField(
            model_name='collectionversionsignature',
            old_name='new_data',
            new_name='data',
        ),
    ]



"""Tests related to upload of collections."""

import hashlib
import logging
import os
import re
import shutil
from datetime import datetime
from urllib.parse import urljoin

import pytest

from pulp_smash import api, config
from pulp_smash.pulp3.utils import gen_distribution, gen_repo
from pulp_smash.utils import http_get

from pulp_ansible.tests.functional.constants import (
    ANSIBLE_COLLECTION_FILE_NAME,
    ANSIBLE_COLLECTION_UPLOAD_FIXTURE_URL,
    ANSIBLE_DISTRIBUTION_PATH,
    ANSIBLE_REPO_PATH,
)
from orionutils.generator import build_collection


logger = logging.getLogger(__name__)


def upload_handler(client, response):
    """Handle responses to collection upload by fetching and returning the task data."""
    response.raise_for_status()
    logger.debug("response status: %s", response.status_code)
    if response.status_code == 204:
        return response
    api._handle_202(client._cfg, response, client.pulp_host)
    if response.request.method == "POST":
        task_url = response.json()["task"]
        task = client.get(task_url)
        return task
    else:
        return response.json()


def get_galaxy_url(base, path):
    """Given an Ansible Distribution base_path and an endpoint path, construct a URL.

    Takes the expected GALAXY_API_ROOT setting of the target into consideration.
    """
    cfg = config.get_config()
    path = path.lstrip("/")
    GALAXY_API_ROOT = cfg.custom.get(
        "galaxy_api_root", "/pulp_ansible/galaxy/%(base_path)s/api/"
    ) % {"base_path": base}
    return urljoin(GALAXY_API_ROOT, path)


@pytest.fixture(scope="session")
def collection_artifact():
    """Generate a randomized collection for testing."""
    # build_collection will only store one collection, so copy to new location and delete later
    artifact = build_collection("skeleton")
    artifact.filename = shutil.copy(artifact.filename, "/tmp")
    yield artifact
    os.remove(artifact.filename)


@pytest.fixture(scope="session")
def collection_artifact2():
    """
    Generate a second randomized collection for testing.

    This collection will have the same namespace and version, but different name.
    """
    artifact2 = build_collection("skeleton")
    return artifact2


def get_metadata_published(pulp_client, pulp_dist):
    """Return published datetime."""
    metadata = pulp_client.using_handler(api.json_handler).get(
        get_galaxy_url(pulp_dist["base_path"], "/v3/")
    )
    return datetime.strptime(metadata["published"], "%Y-%m-%dT%H:%M:%S.%fZ")


def upload_collection(client, filename, base_path):
    """Helper to upload collections to pulp_ansible/galaxy."""
    UPLOAD_PATH = get_galaxy_url(base_path, "/v3/artifacts/collections/")
    collection = {"file": (open(filename, "rb"))}

    return client.using_handler(upload_handler).post(UPLOAD_PATH, files=collection)


@pytest.fixture(scope="session")
def collection_upload(pulp_client, collection_artifact, pulp_dist):
    """Publish a new collection and return the processed response data."""
    published_before_upload = get_metadata_published(pulp_client, pulp_dist)
    response = upload_collection(pulp_client, collection_artifact.filename, pulp_dist["base_path"])
    published_after_upload = get_metadata_published(pulp_client, pulp_dist)
    assert published_after_upload > published_before_upload
    return response


@pytest.fixture(scope="session")
def collection_upload2(pulp_client, collection_artifact2, pulp_dist):
    """Publish the second new collection and return the processed response data."""
    published_before_upload = get_metadata_published(pulp_client, pulp_dist)
    response = upload_collection(pulp_client, collection_artifact2.filename, pulp_dist["base_path"])
    published_after_upload = get_metadata_published(pulp_client, pulp_dist)
    assert published_after_upload > published_before_upload
    return response


@pytest.fixture(scope="session")
def collection_detail(collection_upload, pulp_client, pulp_dist, collection_artifact):
    """Fetch and parse a collection details response from an uploaded collection."""
    url = get_galaxy_url(
        pulp_dist["base_path"],
        f"/v3/collections/{collection_artifact.namespace}/{collection_artifact.name}/",
    )
    response = pulp_client.using_handler(api.json_handler).get(url)
    return response


@pytest.fixture(scope="session")
def pulp_client():
    """Create and configure a Pulp API client, including custom authentication headers."""
    cfg = config.get_config()
    client = api.Client(cfg)
    headers = cfg.custom.get("headers", None)
    if headers:
        client.request_kwargs.setdefault("headers", {}).update(headers)
    return client


@pytest.fixture(scope="session")
def pulp_repo(pulp_client):
    """Find or create a Repository to attach to the Ansible Distribution we create."""
    repos = pulp_client.get(ANSIBLE_REPO_PATH)
    if repos:
        yield repos[0]
    else:
        repo_data = gen_repo(name="automation-hub")
        repo = pulp_client.post(ANSIBLE_REPO_PATH, repo_data)
        yield repo
        pulp_client.delete(repo["pulp_href"])


@pytest.fixture(scope="session")
def pulp_dist(pulp_client, pulp_repo):
    """Create an Ansible Distribution to simulate the automation hub environment for testing."""
    dists = pulp_client.get(ANSIBLE_DISTRIBUTION_PATH + "?base_path=automation-hub")

    if len(dists) == 0:
        dist_data = gen_distribution(
            name="automation-hub", base_path="automation-hub", repository=pulp_repo["pulp_href"]
        )
        dist = pulp_client.post(ANSIBLE_DISTRIBUTION_PATH, dist_data)
        created = True
    elif len(dists) == 1:
        dist = dists[0]
        created = False
    else:
        raise ValueError("Found too many Ansible Distributions at 'automation-hub'.")
    yield dist
    if created:
        pulp_client.delete(dist["pulp_href"])


@pytest.fixture(scope="session")
def known_collection():
    """Fetch and prepare a known collection from Galaxy to use in an upload test."""
    collection_content = http_get(ANSIBLE_COLLECTION_UPLOAD_FIXTURE_URL)
    collection = {"file": (ANSIBLE_COLLECTION_FILE_NAME, collection_content)}
    return collection


def test_collection_upload(collection_upload):
    """Upload a new collection.

    Uploads a newly generated collection and validates the resulting collection version details.
    """
    # Validate the upload response
    assert collection_upload["error"] is None
    assert re.match(
        r"[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}", collection_upload["id"]
    )
    assert collection_upload["state"] == "completed"

    assert "updated_at" in collection_upload
    assert "started_at" in collection_upload
    assert "created_at" in collection_upload
    assert "finished_at" in collection_upload
    assert "messages" in collection_upload

    # TODO: Add this back when namespace, name, and version are apart of the CollectionImport
    # for key, value in collection_upload.items():
    #     if key in COLLECTION_METADATA.keys():
    #         assert COLLECTION_METADATA[key] == value, collection_upload


def test_collection_list(
    collection_artifact,
    collection_artifact2,
    collection_upload,
    collection_upload2,
    pulp_client,
    pulp_dist,
):
    """Tests the collection list endpoint after uploading both collections."""
    url = get_galaxy_url(pulp_dist["base_path"], "v3/collections/")
    response = pulp_client.using_handler(api.json_handler).get(url)

    assert response["meta"]["count"] >= 2
    present_collections = {c["href"].split("collections/")[1] for c in response["data"]}
    uploaded_collections = {
        f"index/{collection_artifact.namespace}/{collection_artifact.name}/",
        f"index/{collection_artifact2.namespace}/{collection_artifact2.name}/",
    }
    assert uploaded_collections.issubset(present_collections)


def test_collection_detail(collection_artifact, collection_detail, pulp_dist):
    """Test collection detail resulting from a successful upload of one version.

    Includes information of the most current version.
    """
    url = (
        f"plugin/ansible/content/{pulp_dist['base_path']}"
        f"/collections/index/{collection_artifact.namespace}/{collection_artifact.name}/"
    )

    assert not collection_detail["deprecated"]

    # Check that the URL ends with the correct path so that this test doesn't fail
    # when galaxy_ng is installed
    assert collection_detail["href"].endswith(url)
    assert collection_detail["namespace"] == collection_artifact.namespace
    assert collection_detail["name"] == collection_artifact.name
    assert collection_detail["highest_version"]["version"] == "1.0.0"
    assert collection_detail["download_count"] == 0


def test_collection_version_list(
    collection_artifact, pulp_client, collection_detail, collection_upload2
):
    """Test the versions endpoint, listing the available versions of a given collection."""
    # Version List Endpoint
    versions = pulp_client.using_handler(api.json_handler).get(collection_detail["versions_url"])
    assert versions["meta"]["count"] == 1
    version = versions["data"][0]

    assert version["version"] == "1.0.0"
    assert version["href"] == collection_detail["highest_version"]["href"]


def test_collection_version_filter_by_q(
    ansible_bindings, pulp_client, pulp_dist, monitor_task, delete_orphans_pre
):
    """Verify successive imports do not aggregate tags into search vectors."""

    def publish(new_artifact):
        body = {
            "file": new_artifact.filename,
            "expected_namespace": new_artifact.namespace,
            "expected_name": new_artifact.name,
            "expected_version": new_artifact.version,
        }
        resp = ansible_bindings.ContentCollectionVersionsApi.create(**body)
        monitor_task(resp.task)

    # make&publish 2 collections with each having unique tags
    specs = [("tag1", "ns1", "col1"), ("tag2", "ns1", "col2")]
    for spec in specs:
        cfg = {
            "namespace": spec[1],
            "name": spec[2],
            "description": "",
            "repository": f"https://github.com/{spec[1]}/{spec[2]}",
            "authors": ["jimbob"],
            "version": "1.0.0",
            "tags": [spec[0]],
        }
        this_artifact = build_collection("skeleton", config=cfg)
        publish(this_artifact)

    for spec in specs:
        resp = ansible_bindings.ContentCollectionVersionsApi.list(q=spec[0])

        # should only get the 1 cv as a result ...
        assert resp.count == 1
        assert resp.results[0].namespace == spec[1]
        assert resp.results[0].name == spec[2]


def test_collection_version(collection_artifact, pulp_client, collection_detail):
    """Test collection version endpoint.

    Each collection version details a specific uploaded artifact for the collection.
    """
    # Version Endpoint
    version = pulp_client.using_handler(api.json_handler).get(
        collection_detail["highest_version"]["href"]
    )

    assert version["name"] == collection_artifact.name
    assert version["namespace"] == {"metadata_sha256": None, "name": collection_artifact.namespace}
    assert version["version"] == "1.0.0"

    tarball = open(collection_artifact.filename, "rb").read()
    assert version["artifact"]["sha256"] == hashlib.sha256(tarball).hexdigest()
    assert version["artifact"]["size"] == len(tarball)

    assert version["artifact"]["filename"] == collection_artifact.filename.strip("/tmp/")

    assert "updated_at" in version
    assert "created_at" in version

    assert "files" in version
    assert "manifest" in version
    assert "requires_ansible" in version

    #     # TODO: Test meta data
    #     # 'metadata': {'authors': ['Orion User 1'],
    #     #             'contents': [],
    #     #             'dependencies': {},
    #     #             'description': 'a collection with some deps on other collections',
    #     #             'documentation': '',
    #     #             'homepage': '',
    #     #             'issues': '',
    #     #             'license': ['GPL-3.0-or-later'],
    #     #             'repository': 'http://github.example.com/orionuser1/skeleton',
    #     #             'tags': ['collectiontest']},


def test_collection_download(collection_artifact, pulp_client, collection_detail):
    """Test collection download URL.

    Should require authentication and redirect to a download location.
    """
    version = pulp_client.using_handler(api.json_handler).get(
        collection_detail["highest_version"]["href"]
    )

    # Artifact Download Endoint
    url = version["download_url"]

    tarball = open(collection_artifact.filename, "rb").read()

    c = pulp_client.using_handler(api.echo_handler)
    f = c.get(url)
    assert f.status_code == 200, (url, f.request.headers)
    assert f.content == tarball


def test_collection_upload_repeat(pulp_client, collection_artifact, pulp_dist, collection_upload):
    """
    Upload a duplicate collection.
    """
    response = upload_collection(pulp_client, collection_artifact.filename, pulp_dist["base_path"])
    assert response["error"] is None
    assert response["state"] == "completed"
    assert re.match(r"[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}", response["id"])



import asyncio
import logging
import math

from asyncio import FIRST_COMPLETED
from gettext import gettext as _

from pulpcore.plugin.models import Artifact, ProgressReport, Remote
from pulpcore.plugin.stages import (
    DeclarativeArtifact,
    DeclarativeContent,
    DeclarativeVersion,
    Stage,
)
from pulp_ansible.app.constants import PAGE_SIZE
from pulp_ansible.app.models import AnsibleRepository, RoleRemote, Role
from pulp_ansible.app.tasks.utils import get_api_version, get_page_url, parse_metadata


log = logging.getLogger(__name__)


# The Github URL template to fetch a .tar.gz file from
GITHUB_URL = "https://github.com/%s/%s/archive/%s.tar.gz"


def synchronize(remote_pk, repository_pk, mirror=False):
    """
    Sync content from the remote repository.

    Create a new version of the repository that is synchronized with the remote.

    Args:
        remote_pk (str): The remote PK.
        repository_pk (str): The repository PK.
        mirror (bool): True for mirror mode, False for additive.

    Raises:
        ValueError: If the remote does not specify a URL to sync.

    """
    remote = RoleRemote.objects.get(pk=remote_pk)
    repository = AnsibleRepository.objects.get(pk=repository_pk)

    if not remote.url:
        raise ValueError(_("A remote must have a url specified to synchronize."))

    log.info(
        _("Synchronizing: repository=%(r)s remote=%(p)s"), {"r": repository.name, "p": remote.name}
    )
    first_stage = RoleFirstStage(remote)
    d_version = DeclarativeVersion(first_stage, repository, mirror=mirror)
    return d_version.create()


class RoleFirstStage(Stage):
    """
    The first stage of a pulp_ansible sync pipeline for roles.
    """

    def __init__(self, remote):
        """
        The first stage of a pulp_ansible sync pipeline.

        Args:
            remote (RoleRemote): The remote data to be used when syncing

        """
        super().__init__()
        self.remote = remote

        # Interpret download policy
        self.deferred_download = self.remote.policy != Remote.IMMEDIATE

    async def run(self):
        """
        Build and emit `DeclarativeContent` from the ansible metadata.
        """
        async with ProgressReport(
            message="Parsing Role Metadata", code="sync.parsing.metadata"
        ) as pb:
            async for metadata in self._fetch_roles():
                for version in metadata["summary_fields"]["versions"]:
                    url = GITHUB_URL % (
                        metadata["github_user"],
                        metadata["github_repo"],
                        version["name"],
                    )
                    role = Role(
                        version=version["name"],
                        name=metadata["name"],
                        namespace=metadata["namespace"],
                    )
                    relative_path = "%s/%s/%s.tar.gz" % (
                        metadata["namespace"],
                        metadata["name"],
                        version["name"],
                    )
                    d_artifact = DeclarativeArtifact(
                        artifact=Artifact(),
                        url=url,
                        relative_path=relative_path,
                        remote=self.remote,
                        deferred_download=self.deferred_download,
                    )
                    d_content = DeclarativeContent(content=role, d_artifacts=[d_artifact])
                    await pb.aincrement()
                    await self.put(d_content)

    async def _fetch_roles(self):
        async for metadata in self._fetch_galaxy_pages():
            for result in metadata["results"]:
                role = {
                    "name": result["name"],
                    "namespace": result["summary_fields"]["namespace"]["name"],
                    "summary_fields": result["summary_fields"],  # needed for versions
                    "github_user": result["github_user"],
                    "github_repo": result["github_repo"],
                }
                yield role

    async def _fetch_galaxy_pages(self):
        """
        Fetch the roles in a remote repository.

        Returns:
            async generator: dicts that represent pages from galaxy api

        """
        page_count = 0
        remote = self.remote

        progress_data = dict(
            message="Parsing Pages from Galaxy Roles API", code="sync.parsing.roles"
        )
        async with ProgressReport(**progress_data) as progress_bar:
            api_version = get_api_version(remote.url)
            downloader = remote.get_downloader(url=get_page_url(remote.url, api_version))
            metadata = parse_metadata(await downloader.run())

            page_count = math.ceil(float(metadata["count"]) / float(PAGE_SIZE))
            progress_bar.total = page_count
            await progress_bar.asave()

            yield metadata
            await progress_bar.aincrement()

            # Concurrent downloads are limited by aiohttp...
            not_done = set(
                remote.get_downloader(url=get_page_url(remote.url, api_version, page)).run()
                for page in range(2, page_count + 1)
            )

            while not_done:
                done, not_done = await asyncio.wait(not_done, return_when=FIRST_COMPLETED)
                for item in done:
                    yield parse_metadata(item.result())
                    await progress_bar.aincrement()



# Generated by Django 2.2.24 on 2021-07-07 17:46

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0068_add_timestamp_of_interest'),
        ('ansible', '0035_deprecation_content'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='ansiblecollectiondeprecated',
            unique_together={('namespace', 'name')},
        ),
        migrations.RemoveField(model_name="ansiblecollectiondeprecated", name="collection_id"),
        migrations.RemoveField(model_name="ansiblecollectiondeprecated", name="repository_id"),
        migrations.RemoveField(model_name="ansiblecollectiondeprecated", name="version_added_id"),
    ]



# Generated by Django 2.2.6 on 2019-10-03 15:05

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0012_auto_20190906_2253'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='collectionimport',
            options={'ordering': ['task__pulp_created']},
        ),
        migrations.RenameField(
            model_name='collection',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='collection',
            old_name='_id',
            new_name='pulp_id',
        ),
        migrations.RenameField(
            model_name='collection',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.RenameField(
            model_name='tag',
            old_name='_created',
            new_name='pulp_created',
        ),
        migrations.RenameField(
            model_name='tag',
            old_name='_last_updated',
            new_name='pulp_last_updated',
        ),
        migrations.AlterField(
            model_name='collectionversion',
            name='collection',
            field=models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE, related_name='versions', to='ansible.Collection', to_field='pulp_id'),
        ),
    ]



# Generated by Django 4.2.1 on 2023-05-31 20:46

from django.db import migrations, models
import django_lifecycle.mixins
import pulpcore.app.models.base


class Migration(migrations.Migration):
    dependencies = [
        ("ansible", "0052_alter_ansiblecollectiondeprecated_content_ptr_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="CollectionDownloadCount",
            fields=[
                (
                    "pulp_id",
                    models.UUIDField(
                        default=pulpcore.app.models.base.pulp_uuid,
                        editable=False,
                        primary_key=True,
                        serialize=False,
                    ),
                ),
                ("pulp_created", models.DateTimeField(auto_now_add=True)),
                ("pulp_last_updated", models.DateTimeField(auto_now=True, null=True)),
                ("namespace", models.CharField(editable=False, max_length=64)),
                ("name", models.CharField(editable=False, max_length=64)),
                ("download_count", models.BigIntegerField(default=0)),
            ],
            options={
                "unique_together": {("namespace", "name")},
            },
            bases=(django_lifecycle.mixins.LifecycleModelMixin, models.Model),
        ),
    ]



"""Utilities for tests for the ansible plugin."""

from urllib.parse import urlparse, parse_qs

from pulp_smash import config
from pulp_smash.pulp3.bindings import delete_orphans, monitor_task, PulpTestCase
from pulp_smash.pulp3.utils import (
    gen_distribution,
    gen_remote,
    gen_repo,
)

from pulp_ansible.tests.functional.constants import (
    ANSIBLE_FIXTURE_URL,
)

from pulpcore.client.pulpcore import (
    ApiClient as CoreApiClient,
    TasksApi,
    StatusApi,
)
from pulpcore.client.pulp_ansible import (
    ApiClient as AnsibleApiClient,
    ContentCollectionVersionsApi,
    DistributionsAnsibleApi,
    PulpAnsibleApiV3CollectionsApi,
    PulpAnsibleApiV3CollectionsVersionsApi,
    RepositoriesAnsibleApi,
    RemotesCollectionApi,
    RemotesGitApi,
    RemotesRoleApi,
    AnsibleRepositorySyncURL,
    RepositoriesAnsibleVersionsApi,
)


cfg = config.get_config()
configuration = cfg.get_bindings_config()


def is_galaxy_ng_installed():
    """Returns whether or not the galaxy_ng plugin is installed."""
    configuration = cfg.get_bindings_config()
    core_client = CoreApiClient(configuration)
    status_client = StatusApi(core_client)

    status = status_client.status_read()

    for plugin in status.versions:
        if plugin.component == "galaxy":
            return True
    return False


def content_counts(repository_version, summary_type="present"):
    content_summary = getattr(repository_version.content_summary, summary_type)
    return {key: value["count"] for key, value in content_summary.items()}


def gen_ansible_client():
    """Return an OBJECT for ansible client."""
    return AnsibleApiClient(configuration)


def gen_ansible_remote(url=ANSIBLE_FIXTURE_URL, include_pulp_auth=False, **kwargs):
    """Return a semi-random dict for use in creating a ansible Remote.

    :param url: The URL of an external content source.
    """
    if include_pulp_auth:
        kwargs["username"] = cfg.pulp_auth[0]
        kwargs["password"] = cfg.pulp_auth[1]

    if "rate_limit" not in kwargs:
        kwargs["rate_limit"] = 5

    return gen_remote(url, **kwargs)


core_client = CoreApiClient(configuration)
tasks = TasksApi(core_client)


class TestCaseUsingBindings(PulpTestCase):
    """A parent TestCase that instantiates the various bindings used throughout tests."""

    @classmethod
    def setUpClass(cls):
        """Create class-wide variables."""
        cls.client = gen_ansible_client()
        cls.repo_api = RepositoriesAnsibleApi(cls.client)
        cls.repo_version_api = RepositoriesAnsibleVersionsApi(cls.client)
        cls.remote_collection_api = RemotesCollectionApi(cls.client)
        cls.remote_git_api = RemotesGitApi(cls.client)
        cls.remote_role_api = RemotesRoleApi(cls.client)
        cls.distributions_api = DistributionsAnsibleApi(cls.client)
        cls.cv_api = ContentCollectionVersionsApi(cls.client)
        cls.collections_v3api = PulpAnsibleApiV3CollectionsApi(cls.client)
        cls.collections_versions_v3api = PulpAnsibleApiV3CollectionsVersionsApi(cls.client)

    @classmethod
    def tearDownClass(cls):
        """Clean class-wide variable."""
        delete_orphans()


class SyncHelpersMixin:
    """A common place for sync helper functions."""

    def _create_repo_and_sync_with_remote(self, remote, distribution=False, **repo_kwargs):
        """
        Create a repository and then sync with the provided `remote`.

        Args:
            remote: The remote to be sync with

        Returns:
            repository: The created repository object to be asserted to.
        """
        # Create the repository.
        repo = self.repo_api.create(gen_repo(**repo_kwargs))
        self.addCleanup(self.repo_api.delete, repo.pulp_href)
        if distribution:
            self._create_distribution_from_repo(repo)
        return self._sync_repo(repo, remote=remote.pulp_href)

    def _create_repo_with_attached_remote_and_sync(self, remote, **repo_kwargs):
        """
        Create a repository with the remote attached, and then sync without specifying the `remote`.

        Args:
            remote: The remote to attach to the repository

        Returns:
            repository: The created repository object to be asserted to.
        """
        # Create the repository.
        repo = self.repo_api.create(gen_repo(remote=remote.pulp_href, **repo_kwargs))
        self.addCleanup(self.repo_api.delete, repo.pulp_href)
        return self._sync_repo(repo)

    def _sync_repo(self, repo, **kwargs):
        """
        Sync the repo with optional `kwarg` parameters passed on to the sync method.

        Args:
            repo: The repository to sync

        Returns:
            repository: The updated repository after the sync is complete
        """
        repository_sync_data = AnsibleRepositorySyncURL(**kwargs)
        sync_response = self.repo_api.sync(repo.pulp_href, repository_sync_data)
        monitor_task(sync_response.task)
        repo = self.repo_api.read(repo.pulp_href)
        return repo

    def _create_distribution_from_repo(self, repo, cleanup=True):
        """
        Create an `AnsibleDistribution` serving the `repo` with the `base_path`.

        Args:
            repo: The repository to serve with the `AnsibleDistribution`
            cleanup: Whether the distribution should be cleaned up

        Returns:
            The created `AnsibleDistribution`.
        """
        # Create a distribution.
        body = gen_distribution()
        body["repository"] = repo.pulp_href
        distribution_create = self.distributions_api.create(body)
        created_resources = monitor_task(distribution_create.task).created_resources
        distribution = self.distributions_api.read(created_resources[0])
        if cleanup:
            self.addCleanup(self.distributions_api.delete, distribution.pulp_href)
        return distribution

    def _create_empty_repo_and_distribution(self, cleanup=True):
        """
        Creates an empty `AnsibleRepository` and an `AnsibleDistribution` serving that repository.

        Args:
            cleanup: Whether the repository and distribution should be cleaned up

        Returns:
            Tuple of the created `AnsibleRepository`, `AnsibleDistribution`
        """
        repo = self.repo_api.create(gen_repo())
        if cleanup:
            self.addCleanup(self.repo_api.delete, repo.pulp_href)
        return repo, self._create_distribution_from_repo(repo, cleanup=cleanup)


def iterate_all(list_func, **kwargs):
    """
    Iterate through all of the items on every page in a paginated list view.
    """
    kwargs
    while kwargs is not None:
        response = list_func(**kwargs)

        for x in response.results:
            yield x

        if response.next:
            qs = parse_qs(urlparse(response.next).query)
            for param in ("offset", "limit"):
                if param in qs:
                    kwargs[param] = qs[param][0]
        else:
            kwargs = None



# Generated by Django 3.2.15 on 2022-09-12 14:00

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django_lifecycle.mixins
import uuid


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('core', '0092_alter_upload_options'),
        ('ansible', '0044_alter_collectionremote_token'),
    ]

    operations = [
        migrations.CreateModel(
            name='DownloadLog',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('ip', models.GenericIPAddressField()),
                ('extra_data', models.JSONField(null=True)),
                ('user_agent', models.TextField()),
                ('content_unit', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='download_logs', to='core.content')),
                ('repository', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='download_logs', to='core.repository')),
                ('repository_version', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='download_logs', to='core.repositoryversion')),
                ('user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='download_logs', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
            bases=(django_lifecycle.mixins.LifecycleModelMixin, models.Model),
        ),
    ]



# Generated by Django 2.2.15 on 2020-09-01 13:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0022_URLField_to_CharField'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collectionremote',
            name='requirements_file',
            field=models.TextField(null=True),
        ),
    ]



# Generated by Django 2.2.17 on 2021-01-29 18:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0030_collectionversion_requires_ansible'),
    ]

    operations = [
        migrations.AddField(
            model_name='ansiblerepository',
            name='last_synced_metadata_time',
            field=models.DateTimeField(null=True),
        ),
    ]



import pytest

from pulpcore.tests.functional.utils import PulpTaskError
from pulp_ansible.tests.functional.constants import ANSIBLE_FIXTURE_CONTENT_SUMMARY


@pytest.mark.parallel
def test_role_sync(
    ansible_bindings,
    ansible_repo_factory,
    ansible_role_remote_factory,
    ansible_sync_factory,
):
    repository = ansible_repo_factory()
    remote = ansible_role_remote_factory()
    assert repository.latest_version_href == repository.pulp_href + "versions/0/"

    repository = ansible_sync_factory(repository, remote=remote.pulp_href)
    assert repository.latest_version_href == repository.pulp_href + "versions/1/"
    version = ansible_bindings.RepositoriesAnsibleVersionsApi.read(repository.latest_version_href)
    assert {
        k: v["count"] for k, v in version.content_summary.present.items()
    } == ANSIBLE_FIXTURE_CONTENT_SUMMARY
    assert {
        k: v["count"] for k, v in version.content_summary.added.items()
    } == ANSIBLE_FIXTURE_CONTENT_SUMMARY

    repository = ansible_sync_factory(repository, remote=remote.pulp_href)
    assert repository.latest_version_href == repository.pulp_href + "versions/1/"


@pytest.mark.parallel
def test_role_sync_invalid(ansible_repo_factory, ansible_role_remote_factory, ansible_sync_factory):
    remote = ansible_role_remote_factory(url="http://i-am-an-invalid-url.com/invalid/")

    with pytest.raises(PulpTaskError) as exc_info:
        ansible_sync_factory(remote=remote.pulp_href)
    assert (
        exc_info.value.task.error["description"]
        == "Could not determine API version for: http://i-am-an-invalid-url.com/invalid/"
    )



# Generated by Django 2.2.3 on 2019-08-02 12:00

import django.contrib.postgres.fields
import django.contrib.postgres.fields.jsonb
from django.db import migrations, models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0002_advanced_collections'),
    ]

    operations = [
        migrations.CreateModel(
            name='Tag',
            fields=[
                ('_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('_created', models.DateTimeField(auto_now_add=True)),
                ('_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('name', models.CharField(editable=False, max_length=32, unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='collectionversion',
            name='authors',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=64), default=list, editable=False, size=None),
        ),
        migrations.AddField(
            model_name='collectionversion',
            name='dependencies',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict, editable=False),
        ),
        migrations.AddField(
            model_name='collectionversion',
            name='description',
            field=models.TextField(blank=True, default='', editable=False),
        ),
        migrations.AddField(
            model_name='collectionversion',
            name='docs_blob',
            field=django.contrib.postgres.fields.jsonb.JSONField(default=dict, editable=False),
        ),
        migrations.AddField(
            model_name='collectionversion',
            name='documentation',
            field=models.URLField(blank=True, default='', editable=False, max_length=128),
        ),
        migrations.AddField(
            model_name='collectionversion',
            name='homepage',
            field=models.URLField(blank=True, default='', editable=False, max_length=128),
        ),
        migrations.AddField(
            model_name='collectionversion',
            name='issues',
            field=models.URLField(blank=True, default='', editable=False, max_length=128),
        ),
        migrations.AddField(
            model_name='collectionversion',
            name='license',
            field=django.contrib.postgres.fields.ArrayField(base_field=models.CharField(max_length=32), default=list, editable=False, size=None),
        ),
        migrations.AddField(
            model_name='collectionversion',
            name='name',
            field=models.CharField(default='', editable=False, max_length=32),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='collectionversion',
            name='namespace',
            field=models.CharField(default='', editable=False, max_length=32),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='collectionversion',
            name='repository',
            field=models.URLField(blank=True, default='', editable=False, max_length=128),
        ),
        migrations.AlterField(
            model_name='collectionversion',
            name='version',
            field=models.CharField(editable=False, max_length=32),
        ),
        migrations.AlterUniqueTogether(
            name='collectionversion',
            unique_together={('namespace', 'name', 'version')},
        ),
        migrations.RemoveField(
            model_name='collectionversion',
            name='metadata',
        ),
        migrations.RemoveField(
            model_name='collectionversion',
            name='quality_score',
        ),
        migrations.AddField(
            model_name='collectionversion',
            name='tags',
            field=models.ManyToManyField(editable=False, to='ansible.Tag'),
        ),
    ]



import hashlib
from unittest import mock
from django.conf import settings
from django.core.files.uploadedfile import SimpleUploadedFile
from django.test import TestCase

from pulp_ansible.app.serializers import RoleSerializer

from pulpcore.plugin.models import Artifact


class TestRoleSerializer(TestCase):
    """Test RoleSerializer."""

    def setUp(self):
        """Set up the RoleSerializer tests."""
        rawbin = b"test content"
        self.artifact = Artifact.objects.create(
            size=len(rawbin),
            file=SimpleUploadedFile("test_filename", rawbin),
            **{
                algorithm: hashlib.new(algorithm, rawbin).hexdigest()
                for algorithm in Artifact.DIGEST_FIELDS
            },
        )
        self.data = {
            "artifact": "{}artifacts/{}/".format(settings.V3_API_ROOT, self.artifact.pk),
            "version": "0.1.2.3",
            "name": "test1",
            "namespace": "testns",
        }

    def test_valid_data(self):
        """Test that the RoleSerializer accepts valid data."""
        view = mock.Mock()
        serializer = RoleSerializer(data=self.data, context={"view": view})
        self.assertTrue(serializer.is_valid())

    def test_duplicate_data(self):
        """Test that the RoleSerializer does not accept data."""
        view = mock.Mock()
        serializer = RoleSerializer(data=self.data, context={"view": view})
        self.assertTrue(serializer.is_valid())
        serializer.save()
        serializer = RoleSerializer(data=self.data, context={"view": view})
        self.assertFalse(serializer.is_valid())



# Generated by Django 2.2.19 on 2021-03-11 23:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0031_ansiblerepository_last_synced_metadata_time'),
    ]

    operations = [
        migrations.AddField(
            model_name='collectionremote',
            name='sync_dependencies',
            field=models.BooleanField(default=True),
        ),
    ]



from gettext import gettext as _
import json
import sys

from django.core.management import BaseCommand, CommandError

from pulp_ansible.app.models import DownloadLog


SEPARATOR = "\t"


class Command(BaseCommand):
    """
    Django management command for getting a data dump of deprecated permission.
    """

    help = _("Print the Download log.")

    def add_arguments(self, parser):
        """Add arguments."""
        parser.add_argument("--tabular", action="store_true", help=_("Output table format."))

    def handle(self, *args, **options):
        """Print the contents of the download log."""
        tabular = options.get("tabular", False)

        if tabular:
            try:
                from prettytable import PrettyTable
            except ImportError:
                raise CommandError("'prettytable' package must be installed for tabular output.")
        qs = DownloadLog.objects.order_by("pulp_created").prefetch_related("content_unit", "user")
        data = [
            {
                "time": str(entry.pulp_created),
                "content_unit": str(entry.content_unit.cast()),
                "user": entry.user and entry.user.username,
                "ip": entry.ip,
                "extra_data": entry.extra_data,
                "user_agent": entry.user_agent,
                "repository": entry.repository.name,
            }
            for entry in qs
        ]

        # User model permissions
        if tabular:
            print("# ==== " + _("Download Log") + " ====")
            table = PrettyTable()
            table.field_names = [
                _("time"),
                _("content unit"),
                _("user"),
                _("ip"),
                _("extra_data"),
                _("user agent"),
                _("repository"),
            ]
            table.add_rows(
                (
                    [
                        entry["time"],
                        entry["content_unit"],
                        entry["user"],
                        entry["ip"],
                        entry["extra_data"],
                        entry["user_agent"],
                        entry["repository"],
                    ]
                    for entry in data
                )
            )
            print(table)
            print()

        else:
            json.dump(
                data,
                sys.stdout,
            )
            print()



"""Tasks for CollectionVersionMark"""

from pulpcore.plugin.tasking import add_and_remove
from pulp_ansible.app.models import AnsibleRepository, CollectionVersion, CollectionVersionMark


def mark(repository_href, content_hrefs, value):
    """The mark task."""
    repository = AnsibleRepository.objects.get(pk=repository_href)
    content = _get_content_queryset(repository, content_hrefs)
    marks_in_repo = _get_list_of_marks_pk_in_repo(repository, content, value)

    marks_to_add = []
    for collection_version in content:
        # A mark for the same collection and value may exist in another repo.
        mark, _created = CollectionVersionMark.objects.get_or_create(
            marked_collection=collection_version, value=value
        )
        if mark.pk not in marks_in_repo:
            marks_to_add.append(mark.pk)

    add_and_remove(
        repository_pk=repository.pk,
        add_content_units=marks_to_add,
        remove_content_units=[],
    )


def unmark(repository_href, content_hrefs, value):
    """The unmark task."""
    repository = AnsibleRepository.objects.get(pk=repository_href)
    content = _get_content_queryset(repository, content_hrefs)
    marks_in_repo = _get_list_of_marks_pk_in_repo(repository, content, value)
    add_and_remove(
        repository_pk=repository.pk,
        add_content_units=[],
        remove_content_units=marks_in_repo,
    )


def _get_list_of_marks_pk_in_repo(repository, content, value):
    """Get all the marks from repo having value and content"""
    return (
        repository.latest_version()
        .content.filter(
            pulp_type=CollectionVersionMark.get_pulp_type(),
            ansible_collectionversionmark__value=value,
            ansible_collectionversionmark__marked_collection__in=content,
        )
        .values_list("pk", flat=True)
    )


def _get_content_queryset(repository, content_hrefs):
    """Return content queryset based on a repo and content hrefs"""
    if content_hrefs == ["*"]:
        content_in_latest_version = repository.latest_version().content.filter(
            pulp_type=CollectionVersion.get_pulp_type()
        )
        return CollectionVersion.objects.filter(pk__in=content_in_latest_version)
    return CollectionVersion.objects.filter(pk__in=content_hrefs)



"""Tests collection sync functionality that is common to both Galaxy V2 and V3."""

import datetime

import pytest

from pulp_ansible.tests.functional.utils import (
    gen_ansible_remote,
    SyncHelpersMixin,
    TestCaseUsingBindings,
)
from pulp_ansible.tests.functional.constants import TEST_COLLECTION_CONFIGS
from orionutils.generator import build_collection
from pulpcore.client.pulp_ansible import PulpAnsibleArtifactsCollectionsV3Api
from pulp_ansible.tests.functional.utils import monitor_task


@pytest.mark.parallel
def test_sync_supports_mirror_option_true(
    ansible_bindings, ansible_collection_remote_factory, ansible_sync_factory
):
    """Sync multiple remotes into the same repo with mirror as `True`."""
    remote_a = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - robertdebock.ansible_development_environment",
        sync_dependencies=False,
    )

    remote_b = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - testing.k8s_demo_collection",
        sync_dependencies=False,
    )

    repository = ansible_sync_factory(remote=remote_a.pulp_href)
    assert repository.latest_version_href == f"{repository.versions_href}1/"
    repository = ansible_sync_factory(repository, remote=remote_b.pulp_href, mirror=True)
    assert repository.latest_version_href == f"{repository.versions_href}2/"

    # Assert more CollectionVersion are present in the first sync than the second
    if repository.retain_repo_versions and repository.retain_repo_versions > 1:
        content_version_one = ansible_bindings.ContentCollectionVersionsApi.list(
            repository_version=f"{repository.pulp_href}versions/1/"
        )
        assert content_version_one.count >= 3
    content_version_two = ansible_bindings.ContentCollectionVersionsApi.list(
        repository_version=f"{repository.pulp_href}versions/2/"
    )
    assert content_version_two.count == 1


@pytest.mark.parallel
def test_sync_supports_mirror_option_false(
    ansible_bindings, ansible_collection_remote_factory, ansible_sync_factory
):
    """Sync multiple remotes into the same repo with mirror as `False`."""
    remote_a = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - robertdebock.ansible_development_environment",
        sync_dependencies=False,
    )

    remote_b = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - testing.k8s_demo_collection",
        sync_dependencies=False,
    )

    repository = ansible_sync_factory(remote=remote_a.pulp_href)
    assert repository.latest_version_href == f"{repository.versions_href}1/"
    repository = ansible_sync_factory(repository, remote=remote_b.pulp_href, mirror=False)
    assert repository.latest_version_href == f"{repository.versions_href}2/"

    # Assert more CollectionVersion are present in the second sync than the first
    if repository.retain_repo_versions and repository.retain_repo_versions > 1:
        content_version_one = ansible_bindings.ContentCollectionVersionsApi.list(
            repository_version=f"{repository.pulp_href}versions/1/"
        )
        assert content_version_one.count >= 3
    content_version_two = ansible_bindings.ContentCollectionVersionsApi.list(
        repository_version=f"{repository.pulp_href}versions/2/"
    )
    assert content_version_two.count == 4


def test_sync_mirror_defaults_to_false(
    ansible_bindings, ansible_collection_remote_factory, ansible_sync_factory
):
    """Sync multiple remotes into the same repo to ensure mirror defaults to `False`."""
    remote_a = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - robertdebock.ansible_development_environment",
        sync_dependencies=False,
    )

    remote_b = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - testing.k8s_demo_collection",
        sync_dependencies=False,
    )

    repository = ansible_sync_factory(remote=remote_a.pulp_href)
    assert repository.latest_version_href == f"{repository.versions_href}1/"
    repository = ansible_sync_factory(repository, remote=remote_b.pulp_href)
    assert repository.latest_version_href == f"{repository.versions_href}2/"

    # Assert more CollectionVersion are present in the second sync than the first
    if repository.retain_repo_versions and repository.retain_repo_versions > 1:
        content_version_one = ansible_bindings.ContentCollectionVersionsApi.list(
            repository_version=f"{repository.pulp_href}versions/1/"
        )
        assert content_version_one.count >= 3
    content_version_two = ansible_bindings.ContentCollectionVersionsApi.list(
        repository_version=f"{repository.pulp_href}versions/2/"
    )
    assert content_version_two.count == 4


@pytest.mark.parallel
@pytest.mark.parametrize(
    "collection,old_galaxy,sync_dependencies,min_count",
    [
        pytest.param("ibm.ibm_zos_core", False, False, 14, id="with_long_tag"),
        pytest.param("rshad.collection_demo", False, False, 6, id="with_dot_slash_in_manifest"),
        pytest.param("brightcomputing.bcm", True, False, 5, id="with_strange_version_numbers"),
        pytest.param(
            "pulp.pulp_installer",
            False,
            True,
            5,
            id="simple_dependency",
            marks=[pytest.mark.timeout(1800)],  # TODO This test takes too much time!
        ),
    ],
)
def test_sync_collection_with_specialities(
    ansible_bindings,
    collection,
    old_galaxy,
    sync_dependencies,
    min_count,
    ansible_collection_remote_factory,
    ansible_sync_factory,
):
    """Sync a collection that is known to be special."""
    remote = ansible_collection_remote_factory(
        url=(
            "https://old-galaxy.ansible.com/api/"
            if old_galaxy
            else "https://galaxy.ansible.com/api/"
        ),
        requirements_file=f"collections:\n  - {collection}",
        sync_dependencies=sync_dependencies,
    )

    repository = ansible_sync_factory(remote=remote.pulp_href)

    content = ansible_bindings.ContentCollectionVersionsApi.list(
        repository_version=f"{repository.pulp_href}versions/1/"
    )
    assert content.count >= min_count


class FullDependenciesSync(TestCaseUsingBindings, SyncHelpersMixin):
    """
    Collection sync tests for syncing collections and their dependencies.

    Dependency Trees:
    A               E
    |             /   \
    B            F     G
    |            |     |
    C         H(1-3)   D
    |
    D

    H has 5 versions, no dependencies
    """

    @classmethod
    def setUpClass(cls):
        """Set up class variables."""
        super().setUpClass()
        cls._build_and_publish_collections()

    @classmethod
    def tearDownClass(cls):
        """Tear down class variables."""
        cls.repo_api.delete(cls.repo.pulp_href)
        cls.distributions_api.delete(cls.distro.pulp_href)

    @classmethod
    def _build_and_publish_collections(cls):
        """Builds and publishes the collections to be used in this test."""
        from django.conf import settings

        cls.collections = []
        cls.repo, cls.distro = cls._create_empty_repo_and_distribution(cls(), cleanup=False)

        upload_api = PulpAnsibleArtifactsCollectionsV3Api(cls.client)
        for cfg in TEST_COLLECTION_CONFIGS:
            collection = build_collection("skeleton", config=cfg)
            upload_response = upload_api.create(cls.distro.base_path, collection.filename)
            api_root = settings.API_ROOT
            monitor_task("{}api/v3/tasks/{}/".format(api_root, upload_response.task[-37:-1]))
            cls.collections.append(collection)
        cls.distro.client_url += "api/"

    def test_simple_one_level_dependency(self):
        """Sync test.c which requires test.d."""
        body = gen_ansible_remote(
            url=self.distro.client_url,
            requirements_file="collections:\n  - test.c",
            include_pulp_auth=True,
        )
        remote = self.remote_collection_api.create(body)
        self.addCleanup(self.remote_collection_api.delete, remote.pulp_href)
        repo = self._create_repo_and_sync_with_remote(remote)

        content = self.cv_api.list(repository_version=f"{repo.pulp_href}versions/1/")
        assert content.count == 2

    def test_simple_multi_level_dependency(self):
        """Sync test.a which should get the dependency chain: test.b -> test.c -> test.d."""
        body = gen_ansible_remote(
            url=self.distro.client_url,
            requirements_file="collections:\n  - test.a",
            include_pulp_auth=True,
        )
        remote = self.remote_collection_api.create(body)
        self.addCleanup(self.remote_collection_api.delete, remote.pulp_href)

        repo = self._create_repo_and_sync_with_remote(remote)

        content = self.cv_api.list(repository_version=f"{repo.pulp_href}versions/1/")
        assert content.count == 4

    def test_complex_one_level_dependency(self):
        """Sync test.f which should get 3 versions of test.h."""
        body = gen_ansible_remote(
            url=self.distro.client_url,
            requirements_file="collections:\n  - test.f",
            include_pulp_auth=True,
        )
        remote = self.remote_collection_api.create(body)
        self.addCleanup(self.remote_collection_api.delete, remote.pulp_href)

        repo = self._create_repo_and_sync_with_remote(remote)

        content = self.cv_api.list(repository_version=f"{repo.pulp_href}versions/1/")
        assert content.count == 4

    def test_complex_multi_level_dependency(self):
        """Sync test.e which should get test.f, test.d, test.g and 3 versions of test.h."""
        body = gen_ansible_remote(
            url=self.distro.client_url,
            requirements_file="collections:\n  - test.e",
            include_pulp_auth=True,
        )
        remote = self.remote_collection_api.create(body)
        self.addCleanup(self.remote_collection_api.delete, remote.pulp_href)

        repo = self._create_repo_and_sync_with_remote(remote)

        content = self.cv_api.list(repository_version=f"{repo.pulp_href}versions/1/")
        assert content.count == 7


@pytest.mark.skip("Skipped until fixture metadata has a published date")
@pytest.mark.parallel
def test_optimized_sync(
    ansible_bindings,
    ansible_repo_factory,
    ansible_collection_remote_factory,
):
    # TODO this test is incomplete and may not work
    remote1 = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - robertdebock.ansible_development_environment",
        sync_dependencies=False,
        signed_only=False,
    )
    remote2 = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - robertdebock.ansible_development_environment",
        sync_dependencies=False,
        signed_only=False,
    )
    repository = ansible_repo_factory(remote=remote1.pulp_href)
    monitor_task(ansible_bindings.RepositoriesAnsibleApi.sync(repository.pulp_href, {}).task)
    repository = ansible_bindings.RepositoriesAnsibleApi.read(repository.pulp_href)
    assert repository.last_synced_metadata_time is not None
    monitor_task(
        ansible_bindings.RepositoriesAnsibleApi.sync(repository.pulp_href, {"optimize": True}).task
    )
    # TODO CHECK IF SYNC WAS OPTIMIZED
    monitor_task(
        ansible_bindings.RepositoriesAnsibleApi.sync(
            repository.pulp_href, {"remote": remote2.pulp_href, "optimize": True}
        ).task
    )
    repository = ansible_bindings.RepositoriesAnsibleApi.read(repository.pulp_href)
    assert repository.last_synced_metadata_time is None
    # TODO CHECK IF SYNC WAS NOT OPTIMIZED


@pytest.mark.parallel
def test_semver_sync(
    ansible_bindings,
    ansible_repo_factory,
    ansible_collection_remote_factory,
):
    remote = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - anil_cm.terraform_provider",
        sync_dependencies=False,
        signed_only=False,
    )
    repository = ansible_repo_factory(remote=remote.pulp_href)
    monitor_task(ansible_bindings.RepositoriesAnsibleApi.sync(repository.pulp_href, {}).task)
    repository = ansible_bindings.RepositoriesAnsibleApi.read(repository.pulp_href)
    content = ansible_bindings.ContentCollectionVersionsApi.list(
        repository_version=repository.latest_version_href
    )
    versions = {item.version for item in content.results}
    # If this fails check that it is still upstream!
    # TODO create a better local fixture to sync from.
    assert "0.0.1-rerelease+meta" in versions
    assert "0.0.0-rerelease+meta" in versions


@pytest.mark.parallel
def test_last_synced_metadata_time(
    ansible_bindings,
    ansible_repo_factory,
    ansible_collection_remote_factory,
):
    remote1 = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file="collections:\n  - robertdebock.ansible_development_environment",
        sync_dependencies=False,
        signed_only=False,
    )
    repository = ansible_repo_factory(remote=remote1.pulp_href)
    monitor_task(
        ansible_bindings.RepositoriesAnsibleApi.partial_update(
            repository.pulp_href, {"last_synced_metadata_time": "2000-01-01 00:00"}
        ).task
    )
    repository = ansible_bindings.RepositoriesAnsibleApi.read(repository.pulp_href)
    assert repository.last_synced_metadata_time == datetime.datetime(
        2000, 1, 1, 0, 0, tzinfo=datetime.timezone.utc
    )
    monitor_task(
        ansible_bindings.RemotesCollectionApi.partial_update(
            remote1.pulp_href, {"signed_only": True}
        ).task
    )
    repository = ansible_bindings.RepositoriesAnsibleApi.read(repository.pulp_href)
    assert repository.last_synced_metadata_time is None



from django.contrib.postgres import search as psql_search
from django.db import migrations


# This query generates full text search index based
# the following data ranked from A to D:
#   - Namespace name (weight A)
#   - Collection name (weight A)
#   - Collection tags (weight B)
#   - Collection content names (weight C)
#   - Collection description (weight D)
TS_VECTOR_SELECT = '''
    setweight(to_tsvector(coalesce(namespace,'')), 'A')
    || setweight(to_tsvector(coalesce(name, '')), 'A')
    || (
      SELECT
        setweight(to_tsvector(
          coalesce(string_agg("ansible_tag"."name", ' '), '')
        ), 'B')
      FROM
        "ansible_tag" INNER JOIN "ansible_collectionversion_tags"
        ON
        ("ansible_tag"."pulp_id" = "ansible_collectionversion_tags"."tag_id" )
      WHERE
        "ansible_collectionversion_tags"."collectionversion_id" = cv.content_ptr_id
    )
    || (
      SELECT
        setweight(to_tsvector(
          coalesce(string_agg(cvc ->> 'name', ' '), '')
        ), 'C')
      FROM jsonb_array_elements(cv.contents) AS cvc
    )
    || setweight(to_tsvector(coalesce(description, '')), 'D')
'''

# Forces the trigger to rebuild all search_vector values
REBUILD_COLLECTIONS_TS_VECTOR = f'''
UPDATE ansible_collectionversion SET is_highest = False where is_highest = False;
UPDATE ansible_collectionversion SET is_highest = True where is_highest = True;
'''


# Creates a database function and a trigger to update collection search
# vector field when a collection reference to a newer version is updated.
#
# Since it's not possible to insert a collection version before a collection, a latest_version_id
# always gets updated as a separated query after collectionversion is inserted. Thus only `ON
# UPDATE` trigger is required.
CREATE_COLLECTIONS_TS_VECTOR_TRIGGER = f'''
CREATE OR REPLACE FUNCTION update_collection_ts_vector()
    RETURNS TRIGGER AS
$$
BEGIN
    NEW.search_vector := (
        SELECT {TS_VECTOR_SELECT}
        FROM ansible_collectionversion cv
        WHERE cv.content_ptr_id = NEW.content_ptr_id
    );
    RETURN NEW;
END;
$$ LANGUAGE plpgsql;
DROP TRIGGER IF EXISTS update_ts_vector ON ansible_collectionversion;
CREATE TRIGGER update_ts_vector
    BEFORE UPDATE
    ON ansible_collectionversion
    FOR EACH ROW
EXECUTE PROCEDURE update_collection_ts_vector();
'''


DROP_COLLECTIONS_TS_VECTOR_TRIGGER = '''
DROP TRIGGER IF EXISTS update_ts_vector ON ansible_collectionversion;
DROP FUNCTION IF EXISTS update_collection_ts_vector();
'''


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0045_downloadlog'),
    ]

    operations = [
        migrations.RunSQL(
            sql=CREATE_COLLECTIONS_TS_VECTOR_TRIGGER,
            reverse_sql=DROP_COLLECTIONS_TS_VECTOR_TRIGGER,
        ),
        migrations.RunSQL(
            sql=REBUILD_COLLECTIONS_TS_VECTOR,
            reverse_sql=migrations.RunSQL.noop,
        )
    ]



from django.db.models import Q, OuterRef, Subquery, CharField
from pulpcore.plugin.models import RepositoryVersion, RepositoryContent, Task
from django.db.models.functions import Cast, JSONObject


def filter_content_for_repo_version(qs, repo_version):
    """
    Returns the given queryset, but filtered to only include content
    from the specified repository version.

    qs: Content queryset. Must be a model that is or inherits from
        pulpcore.plugin.models.Content
    repo_version: repository version to return content from

    This generally seems to be faster than repo_version.get_content()
    """

    repo_version_qs = RepositoryVersion.objects.filter(
        repository=repo_version.repository_id, number__lte=repo_version.number
    ).values_list("pk")

    f = (
        Q(repository=repo_version.repository_id)
        & Q(version_added__in=repo_version_qs)
        & Q(Q(version_removed=None) | ~Q(version_removed__in=repo_version_qs))
    )
    content_rel = RepositoryContent.objects.filter(f)

    return qs.filter(pk__in=content_rel.values_list("content_id"))


def get_queryset_annotated_with_last_sync_task(qs):
    last_task = (
        Task.objects.filter(
            name__contains="sync",
            reserved_resources_record__icontains=Cast(OuterRef("pk"), output_field=CharField()),
        )
        .values(
            json=JSONObject(
                pk="pk",
                state="state",
                pulp_created="pulp_created",
                finished_at="finished_at",
                error="error",
            )
        )
        .order_by("-pulp_created")[:1]
    )

    return qs.annotate(last_sync_task=Subquery(last_task))



import requests

from pulpcore.client.pulp_ansible import (
    PulpAnsibleDefaultApiV3PluginAnsibleContentCollectionsIndexApi,
    PulpAnsibleDefaultApiV3PluginAnsibleContentCollectionsIndexVersionsApi,
)


# Using requests for these tests because the client follows redirects and I need to inspect the URL
# that gets redirected to


def _get_download_url(client, distribution, namespace, name):
    collection_api = PulpAnsibleDefaultApiV3PluginAnsibleContentCollectionsIndexApi(client)
    versions_api = PulpAnsibleDefaultApiV3PluginAnsibleContentCollectionsIndexVersionsApi(client)

    collection = collection_api.read(
        distribution.base_path,
        name,
        namespace,
    )

    version = versions_api.read(
        distribution.base_path, name, namespace, collection.highest_version["version"]
    )

    return version.download_url


def test_download(
    skip_on_galaxy,
    bindings_cfg,
    ansible_bindings,
    ansible_repo_factory,
    ansible_distribution_factory,
    build_and_upload_collection,
):
    """Test that downloads without content guards work correctly."""
    # Setup distribution
    repository = ansible_repo_factory()
    collection, collection_href = build_and_upload_collection(repository)
    distribution = ansible_distribution_factory(repository=repository)

    # Fetch content
    download_url = _get_download_url(
        ansible_bindings.client, distribution, collection.namespace, collection.name
    )
    pulp_auth = (bindings_cfg.username, bindings_cfg.password)
    response = requests.get(download_url, auth=pulp_auth, allow_redirects=False)

    # Verify that the download url redirects to the content app
    assert response.is_redirect
    content_app_url = response.headers["Location"]

    # Verify that the collection can be downloaded without authentication
    assert "validate_token" not in content_app_url

    response = requests.get(content_app_url)
    assert response.status_code == 200


def test_download_with_content_guard(
    pulpcore_bindings,
    ansible_bindings,
    bindings_cfg,
    gen_object_with_cleanup,
    ansible_repo_factory,
    ansible_distribution_factory,
    build_and_upload_collection,
):
    """Test that downloads with content guards work correctly."""
    # Setup content guarded distribution
    guard = gen_object_with_cleanup(
        pulpcore_bindings.ContentguardsContentRedirectApi, {"name": "test-content-guard"}
    )
    repository = ansible_repo_factory()
    collection, collection_href = build_and_upload_collection(repository)
    distribution = ansible_distribution_factory(
        repository=repository, content_guard=guard.pulp_href
    )

    # Fetch content
    download_url = _get_download_url(
        ansible_bindings.client, distribution, collection.namespace, collection.name
    )
    pulp_auth = (bindings_cfg.username, bindings_cfg.password)
    response = requests.get(download_url, auth=pulp_auth, allow_redirects=False)

    # Verify that the download url redirects to the content app
    assert response.is_redirect
    content_app_url = response.headers["Location"]

    # Verify that token is present
    assert "validate_token" in content_app_url

    collection = requests.get(content_app_url)
    assert collection.status_code == 200

    # Perform an unauthenticated call to the content app and verify that it gets rejected
    collection = requests.get(content_app_url.split("?")[0])
    assert collection.status_code == 403



import contextlib
import hashlib
import os
import shutil
import subprocess
import tempfile
import yaml

from django.core.files.uploadedfile import SimpleUploadedFile

from pulp_ansible.app.models import Collection
from pulp_ansible.app.models import CollectionVersion
from pulpcore.plugin.models import Artifact
from pulpcore.plugin.models import ContentArtifact


@contextlib.contextmanager
def make_cv_tarball(namespace, name, version):
    """Create a collection version from scratch."""
    tdir = tempfile.mkdtemp()
    subprocess.run(f"ansible-galaxy collection init {namespace}.{name}", shell=True, cwd=tdir)
    os.makedirs(os.path.join(tdir, namespace, name, "meta"), exist_ok=True)
    with open(os.path.join(tdir, namespace, name, "meta", "runtime.yml"), "w") as f:
        f.write('requires_ansible: ">=2.13"\n')
    with open(os.path.join(tdir, namespace, name, "README.md"), "w") as f:
        f.write("# title\ncollection docs\n")
    if version is not None:
        with open(os.path.join(tdir, namespace, name, "galaxy.yml"), "r") as f:
            gdata = yaml.safe_load(f.read())
        gdata["version"] = version
        with open(os.path.join(tdir, namespace, name, "galaxy.yml"), "w") as f:
            f.write(yaml.safe_dump(gdata))
    build_pid = subprocess.run(
        "ansible-galaxy collection build .",
        shell=True,
        cwd=os.path.join(tdir, namespace, name),
        stdout=subprocess.PIPE,
    )
    tarfn = build_pid.stdout.decode("utf-8").strip().split()[-1]
    yield tarfn
    shutil.rmtree(tdir)


def build_cvs_from_specs(specs, build_artifacts=True):
    """Make CVs from a list of [namespace, name, version] specs."""
    collection_versions = []
    for spec in specs:
        col, _ = Collection.objects.get_or_create(name=spec[0])
        col.save()
        cv = CollectionVersion(collection=col, namespace=spec[0], name=spec[1], version=spec[2])
        cv.save()
        collection_versions.append(cv)

        if build_artifacts:
            with make_cv_tarball(spec[0], spec[1], spec[2]) as tarfn:
                rawbin = open(tarfn, "rb").read()
                artifact = Artifact.objects.create(
                    size=os.path.getsize(tarfn),
                    file=SimpleUploadedFile(tarfn, rawbin),
                    **{
                        algorithm: hashlib.new(algorithm, rawbin).hexdigest()
                        for algorithm in Artifact.DIGEST_FIELDS
                    },
                )
                artifact.save()

            ca = ContentArtifact.objects.create(
                artifact=artifact, content=cv, relative_path=cv.relative_path
            )
            ca.save()

    return collection_versions



from gettext import gettext as _

import json

from django.db import transaction
from django.conf import settings
from jsonschema import Draft7Validator
from rest_framework import serializers
from drf_spectacular.utils import extend_schema_field

from galaxy_importer.constants import NAME_REGEXP
from pulpcore.plugin.models import Artifact, ContentArtifact, SigningService
from pulpcore.plugin.serializers import (
    DetailRelatedField,
    ContentChecksumSerializer,
    ModelSerializer,
    NoArtifactContentSerializer,
    NoArtifactContentUploadSerializer,
    RelatedField,
    RemoteSerializer,
    RepositorySerializer,
    RepositorySyncURLSerializer,
    SingleArtifactContentSerializer,
    SingleArtifactContentUploadSerializer,
    DistributionSerializer,
    RepositoryVersionRelatedField,
    validate_unknown_fields,
)
from pulpcore.plugin.util import get_url
from rest_framework.exceptions import ValidationError

from .models import (
    AnsibleDistribution,
    CollectionVersionMark,
    GitRemote,
    RoleRemote,
    AnsibleCollectionDeprecated,
    AnsibleNamespace,
    AnsibleNamespaceMetadata,
    AnsibleRepository,
    Collection,
    CollectionImport,
    CollectionVersion,
    CollectionVersionSignature,
    CollectionRemote,
    Role,
    Tag,
)
from pulp_ansible.app import fields
from pulp_ansible.app.schema import COPY_CONFIG_SCHEMA
from pulp_ansible.app.tasks.utils import (
    parse_collections_requirements_file,
    parse_collection_filename,
)
from pulp_ansible.app.tasks.signature import verify_signature_upload
from pulp_ansible.app.tasks.upload import process_collection_artifact, finish_collection_upload


class RoleSerializer(SingleArtifactContentSerializer):
    """
    A serializer for Role versions.
    """

    name = serializers.CharField()
    namespace = serializers.CharField()
    version = serializers.CharField()

    def validate(self, data):
        """
        Validates data.

        Args:
            data (dict): User data to validate

        Returns:
            dict: Validated data

        Raises:
            rest_framework.serializers.ValidationError: If invalid data

        """
        data = super().validate(data)
        relative_path = "{namespace}/{name}/{version}.tar.gz".format(
            namespace=data["namespace"], name=data["name"], version=data["version"]
        )
        data["relative_path"] = relative_path
        return data

    class Meta:
        fields = tuple(set(SingleArtifactContentSerializer.Meta.fields) - {"relative_path"}) + (
            "version",
            "name",
            "namespace",
        )
        model = Role


class RoleRemoteSerializer(RemoteSerializer):
    """
    A serializer for Ansible Remotes.
    """

    class Meta:
        fields = RemoteSerializer.Meta.fields
        model = RoleRemote


class GitRemoteSerializer(RemoteSerializer):
    """
    A serializer for Git Collection Remotes.
    """

    metadata_only = serializers.BooleanField(
        help_text=_(
            "If True, only metadata about the content will be stored in Pulp. Clients will "
            "retrieve content from the remote URL."
        ),
        required=False,
    )
    git_ref = serializers.CharField(
        help_text=_("A git ref. e.g.: branch, tag, or commit sha."),
        required=False,
    )

    class Meta:
        model = GitRemote
        fields = tuple(set(RemoteSerializer.Meta.fields) - {"policy"}) + (
            "metadata_only",
            "git_ref",
        )


class AnsibleRepositorySerializer(RepositorySerializer):
    """
    Serializer for Ansible Repositories.
    """

    last_synced_metadata_time = serializers.DateTimeField(
        help_text=_("Last synced metadata time."), allow_null=True, required=False
    )
    gpgkey = serializers.CharField(
        help_text="Gpg public key to verify collection signatures against",
        required=False,
        allow_null=True,
    )

    last_sync_task = serializers.SerializerMethodField()

    class Meta:
        fields = RepositorySerializer.Meta.fields + (
            "last_synced_metadata_time",
            "gpgkey",
            "last_sync_task",
            "private",
        )
        model = AnsibleRepository

    def get_last_sync_task(self, obj):
        if hasattr(obj, "last_sync_task"):
            return obj.last_sync_task

        return None


class AnsibleRepositorySyncURLSerializer(RepositorySyncURLSerializer):
    """
    Serializer for Ansible Repository Sync URL.
    """

    optimize = serializers.BooleanField(
        help_text=_("Whether to optimize sync or not."), default=True
    )

    class Meta:
        fields = RepositorySerializer.Meta.fields + ("optimize",)
        model = AnsibleRepository


class AnsibleRepositoryRebuildSerializer(serializers.Serializer):
    """
    Serializer for Ansible Repository Rebuild.
    """

    namespace = serializers.CharField(
        required=False,
        allow_null=True,
    )
    name = serializers.CharField(
        required=False,
        allow_null=True,
    )
    version = serializers.CharField(
        required=False,
        allow_null=True,
    )


class CollectionRemoteSerializer(RemoteSerializer):
    """
    A serializer for Collection Remotes.
    """

    requirements_file = serializers.CharField(
        help_text=_("The string version of Collection requirements yaml."),
        required=False,
        allow_null=True,
    )
    auth_url = serializers.CharField(
        help_text=_("The URL to receive a session token from, e.g. used with Automation Hub."),
        allow_null=True,
        required=False,
        max_length=255,
    )
    token = serializers.CharField(
        help_text=_(
            "The token key to use for authentication. See https://docs.ansible.com/ansible/"
            "latest/user_guide/collections_using.html#configuring-the-ansible-galaxy-client"
            "for more details"
        ),
        allow_null=True,
        required=False,
        max_length=2000,
        write_only=True,
    )

    sync_dependencies = serializers.BooleanField(
        help_text=_("Sync dependencies for collections specified via requirements file"),
        default=True,
    )

    signed_only = serializers.BooleanField(
        help_text=_("Sync only collections that have a signature"),
        default=False,
    )

    last_sync_task = serializers.SerializerMethodField()

    def validate(self, data):
        """
        Validate collection remote data.

            - url: to ensure it ends with slashes.
            - requirements_file: to ensure it is a valid yaml file.

        Args:
            data (dict): User data to validate

        Returns:
            dict: Validated data

        Raises:
            rest_framework.serializers.ValidationError: If the url or requirements_file is invalid

        """

        def _validate_url(url):
            """Based on https://git.io/JTMAA."""
            if not url.endswith("/") and url != "https://galaxy.ansible.com":
                raise serializers.ValidationError(
                    _("Invalid URL {url}. Ensure the URL ends '/'.").format(url=url)
                )

        data = super().validate(data)

        if data.get("url"):
            _validate_url(data["url"])

        if data.get("requirements_file"):
            collections = parse_collections_requirements_file(data["requirements_file"])
            for collection in collections:
                if collection[2]:
                    _validate_url(collection[2])

        has_token = "token" in data or getattr(self.instance, "token", False)
        if data.get("auth_url") and not has_token:
            raise serializers.ValidationError(
                _("When specifying 'auth_url' you must also specify 'token'.")
            )

        return data

    def get_last_sync_task(self, obj):
        if hasattr(obj, "last_sync_task"):
            return obj.last_sync_task

        return None

    class Meta:
        fields = RemoteSerializer.Meta.fields + (
            "requirements_file",
            "auth_url",
            "token",
            "sync_dependencies",
            "signed_only",
            "last_sync_task",
        )
        model = CollectionRemote


class CollectionOneShotSerializer(serializers.Serializer):
    """
    A serializer for the Collection One Shot Upload API.
    """

    file = serializers.FileField(help_text=_("The Collection tarball."), required=True)

    sha256 = serializers.CharField(
        help_text=_("An optional sha256 checksum of the uploaded file."),
        required=False,
        default=None,
    )

    expected_namespace = serializers.CharField(
        help_text=_(
            "The expected 'namespace' of the Collection to be verified against the "
            "metadata during import."
        ),
        required=False,
        default=None,
    )

    expected_name = serializers.CharField(
        help_text=_(
            "The expected 'name' of the Collection to be verified against the metadata during "
            "import."
        ),
        required=False,
        default=None,
    )

    expected_version = serializers.CharField(
        help_text=_(
            "The expected version of the Collection to be verified against the metadata during "
            "import."
        ),
        required=False,
        default=None,
    )


class AnsibleDistributionSerializer(DistributionSerializer):
    """
    Serializer for Ansible Distributions.
    """

    client_url = serializers.SerializerMethodField(
        read_only=True, help_text=_("The URL of a Collection content source.")
    )
    repository_version = RepositoryVersionRelatedField(
        required=False, help_text=_("RepositoryVersion to be served"), allow_null=True
    )

    def get_client_url(self, obj) -> str:
        """
        Get client_url.
        """
        return "{hostname}/pulp_ansible/galaxy/{base_path}/".format(
            hostname=settings.ANSIBLE_API_HOSTNAME, base_path=obj.base_path
        )

    class Meta:
        fields = (
            "pulp_href",
            "pulp_created",
            "base_path",
            "content_guard",
            "name",
            "repository",
            "repository_version",
            "client_url",
            "pulp_labels",
        )
        model = AnsibleDistribution


class TagSerializer(serializers.ModelSerializer):
    """
    A serializer for the Tag model.
    """

    count = serializers.IntegerField(read_only=True)

    class Meta:
        model = Tag
        fields = ["name", "count"]


class TagNestedSerializer(ModelSerializer):
    """
    A serializer for nesting in the CollectionVersion model.
    """

    name = serializers.CharField(help_text=_("The name of the Tag."), read_only=True)

    class Meta:
        model = Tag
        # this serializer is meant to be nested inside CollectionVersion serializer, so it will not
        # have its own endpoint. That's why we need to explicitly define fields to exclude other
        # inherited fields
        fields = ("name",)


class CollectionSerializer(ModelSerializer):
    """
    A serializer for Ansible Collections.
    """

    name = serializers.CharField(help_text=_("The name of the Collection."))
    namespace = serializers.CharField(help_text=_("The namespace of the Collection."))

    class Meta:
        model = Collection
        fields = ("name", "namespace")


class CollectionDeprecatedSerializer(NoArtifactContentSerializer):
    """
    A serializer for Ansible Collection Deprecations.
    """

    name = serializers.CharField(help_text=_("The name of the Collection."))
    namespace = serializers.CharField(help_text=_("The namespace of the Collection."))

    class Meta:
        model = AnsibleCollectionDeprecated
        fields = NoArtifactContentSerializer.Meta.fields + ("name", "namespace")


class CollectionVersionUploadSerializer(SingleArtifactContentUploadSerializer):
    """
    A serializer with the logic necessary to upload a CollectionVersion.

    Used in ``.viewsets.CollectionVersionViewSet`` and ``.galaxy.v3.views.CollectionUploadViewSet``
    to perform the creation and validation of a CollectionVersion and add to repository if
    necessary. This serializer is meant to be compliant with
    ``pulpcore.plugin.viewsets.SingleArtifactContentUploadViewSet`` and thus follows these steps on
    creation:

    1. ``SingleArtifactContentUploadViewSet.create()`` calls ``validate()`` with request in context
    2. Task payload is created, converting uploaded-file (if present) to an artifact
    3. ``pulpcore.plugin.tasks.general_create`` is dispatched with this serializer, the task payload
        and a deferred context determine by the viewset (default an empty context).
    4. ``general_create`` calls ``validate()`` again with deferred context now.
        ``deferred_validate()`` is now called and the upload object (if present) is converted to an
        artifact.
    5. ``general_create`` calls ``save()`` which will call ``create()``. ``create`` uses the
        validated data to create and save the CollectionVersion. If repository is specified the
        CollectionVersion is then added to the repository.
    """

    sha256 = serializers.CharField(
        help_text=_("An optional sha256 checksum of the uploaded file."),
        required=False,
        write_only=True,
    )

    expected_name = serializers.CharField(
        help_text=_("The name of the collection."),
        max_length=64,
        required=False,
        write_only=True,
    )

    expected_namespace = serializers.CharField(
        help_text=_("The namespace of the collection."),
        max_length=64,
        required=False,
        write_only=True,
    )

    expected_version = serializers.CharField(
        help_text=_("The version of the collection."),
        max_length=128,
        required=False,
        write_only=True,
    )

    def validate(self, data):
        """Check and set the namespace, name & version."""
        fields = ("namespace", "name", "version")
        if not all((f"expected_{x}" in data for x in fields)):
            if not ("file" in data or "filename" in self.context):
                raise ValidationError(
                    _(
                        "expected_namespace, expected_name, and expected_version must be "
                        "specified when using artifact or upload objects"
                    )
                )
            filename = self.context.get("filename") or data["file"].name
            try:
                collection = parse_collection_filename(filename)
            except ValueError:
                raise ValidationError(
                    _("Failed to parse Collection file upload '{}'").format(filename)
                )
            data["expected_namespace"] = collection.namespace
            data["expected_name"] = collection.name
            data["expected_version"] = collection.version

        # Super will call deferred_validate on second call in task context
        return super().validate(data)

    def deferred_validate(self, data):
        """Import the CollectionVersion extracting the metadata from its artifact."""
        # Call super to ensure that data contains artifact
        data = super().deferred_validate(data)
        artifact = data.get("artifact")
        if (sha256 := data.pop("sha256", None)) and sha256 != artifact.sha256:
            raise ValidationError(_("Expected sha256 did not match uploaded artifact's sha256"))

        collection_info = process_collection_artifact(
            artifact=artifact,
            namespace=data.pop("expected_namespace"),
            name=data.pop("expected_name"),
            version=data.pop("expected_version"),
        )
        # repository field clashes
        collection_info["origin_repository"] = collection_info.pop("repository", None)
        data.update(collection_info)
        # `retrieve` needs artifact, but it won't be in validated_data because of `get_artifacts`
        self.context["artifact"] = artifact

        return data

    def retrieve(self, validated_data):
        """Reuse existing CollectionVersion if provided artifact matches."""
        namespace = validated_data["namespace"]
        name = validated_data["name"]
        version = validated_data["version"]
        artifact = self.context["artifact"]
        # TODO switch this check to use digest when ColVersion uniqueness constraint is changed
        col = CollectionVersion.objects.filter(
            namespace=namespace, name=name, version=version
        ).first()
        if col:
            if col._artifacts.get() != artifact:
                raise ValidationError(
                    _("Collection {}.{}-{} already exists with a different artifact").format(
                        namespace, name, version
                    )
                )

        return col

    def create(self, validated_data):
        """Final step in creating the CollectionVersion."""
        tags = validated_data.pop("tags")
        origin_repository = validated_data.pop("origin_repository")
        # Create CollectionVersion from its metadata and adds to repository if specified
        content = super().create(validated_data)

        # Now add tags and update latest CollectionVersion
        finish_collection_upload(content, tags=tags, origin_repository=origin_repository)

        return content

    class Meta:
        fields = tuple(
            set(SingleArtifactContentUploadSerializer.Meta.fields) - {"relative_path"}
        ) + (
            "sha256",
            "expected_name",
            "expected_namespace",
            "expected_version",
        )
        model = CollectionVersion


class CollectionVersionSerializer(ContentChecksumSerializer, CollectionVersionUploadSerializer):
    """
    A serializer for CollectionVersion Content.
    """

    id = serializers.UUIDField(source="pk", help_text="A collection identifier.", read_only=True)

    authors = serializers.ListField(
        help_text=_("A list of the CollectionVersion content's authors."),
        child=serializers.CharField(max_length=64),
        read_only=True,
    )

    contents = fields.JSONDictField(
        help_text=_("A JSON field with data about the contents."), read_only=True
    )

    dependencies = fields.JSONDictField(
        help_text=_(
            "A dict declaring Collections that this collection requires to be installed for it to "
            "be usable."
        ),
        read_only=True,
    )

    description = serializers.CharField(
        help_text=_("A short summary description of the collection."),
        allow_blank=True,
        read_only=True,
    )

    docs_blob = fields.JSONDictField(
        help_text=_("A JSON field holding the various documentation blobs in the collection."),
        read_only=True,
    )

    manifest = fields.JSONDictField(
        help_text=_("A JSON field holding MANIFEST.json data."), read_only=True
    )

    files = fields.JSONDictField(
        help_text=_("A JSON field holding FILES.json data."), read_only=True
    )

    documentation = serializers.CharField(
        help_text=_("The URL to any online docs."),
        allow_blank=True,
        max_length=2000,
        read_only=True,
    )

    homepage = serializers.CharField(
        help_text=_("The URL to the homepage of the collection/project."),
        allow_blank=True,
        max_length=2000,
        read_only=True,
    )

    issues = serializers.CharField(
        help_text=_("The URL to the collection issue tracker."),
        allow_blank=True,
        max_length=2000,
        read_only=True,
    )

    license = serializers.ListField(
        help_text=_("A list of licenses for content inside of a collection."),
        child=serializers.CharField(max_length=32),
        read_only=True,
    )

    name = serializers.CharField(
        help_text=_("The name of the collection."), max_length=64, read_only=True
    )

    namespace = serializers.CharField(
        help_text=_("The namespace of the collection."), max_length=64, read_only=True
    )

    origin_repository = serializers.CharField(
        help_text=_("The URL of the originating SCM repository."),
        source="repository",
        allow_blank=True,
        max_length=2000,
        read_only=True,
    )

    tags = TagNestedSerializer(many=True, read_only=True)

    version = serializers.CharField(
        help_text=_("The version of the collection."), max_length=128, read_only=True
    )

    requires_ansible = serializers.CharField(
        help_text=_(
            "The version of Ansible required to use the collection. "
            "Multiple versions can be separated with a comma."
        ),
        allow_null=True,
        read_only=True,
        max_length=255,
    )

    creating = True

    def validate(self, data):
        """Run super() validate if creating, else return data."""
        # This validation is for creating CollectionVersions
        if not self.creating or self.instance:
            return data
        return super().validate(data)

    def is_valid(self, raise_exception=False):
        """
        Allow this serializer to be used for validating before saving a model.

        See Validating Models:
            https://docs.pulpproject.org/pulpcore/plugins/plugin-writer/concepts/index.html
        """
        write_fields = set(CollectionVersionUploadSerializer.Meta.fields) - {
            "pulp_created",
            "pulp_last_updated",
        }
        if hasattr(self, "initial_data"):
            if any((x in self.initial_data for x in self.Meta.read_fields)):
                # Pop shared fields: artifact & repository
                artifact = self.initial_data.pop("artifact", None)
                repository = self.initial_data.pop("repository", None)
                if any((x in self.initial_data for x in write_fields)):
                    if raise_exception:
                        raise ValidationError(
                            _("Read and write fields can not be used at the same time")
                        )
                    return False
                # Only read fields set, change each one from read_only so they are validated
                for name, field in self.fields.items():
                    if name in self.Meta.read_fields:
                        field.read_only = False
                # Put back in shared fields
                if artifact is not None:
                    self.initial_data["artifact"] = artifact
                if repository is not None:
                    self.initial_data["origin_repository"] = repository
                self.creating = False

        return super().is_valid(raise_exception=raise_exception)

    class Meta:
        read_fields = (
            "id",
            "authors",
            "contents",
            "dependencies",
            "description",
            "docs_blob",
            "manifest",
            "files",
            "documentation",
            "homepage",
            "issues",
            "license",
            "name",
            "namespace",
            "origin_repository",
            "tags",
            "version",
            "requires_ansible",
        )
        fields = (
            CollectionVersionUploadSerializer.Meta.fields
            + ContentChecksumSerializer.Meta.fields
            + read_fields
        )
        model = CollectionVersion


class CollectionVersionSignatureSerializer(NoArtifactContentUploadSerializer):
    """
    A serializer for signature models.
    """

    signed_collection = DetailRelatedField(
        help_text=_("The content this signature is pointing to."),
        view_name_pattern=r"content(-.*/.*)-detail",
        queryset=CollectionVersion.objects.all(),
    )
    pubkey_fingerprint = serializers.CharField(
        help_text=_("The fingerprint of the public key."),
        read_only=True,
    )
    signing_service = RelatedField(
        help_text=_("The signing service used to create the signature."),
        view_name="signing-services-detail",
        read_only=True,
        allow_null=True,
    )

    def deferred_validate(self, data):
        """
        Verify the signature is valid before creating it.
        """
        data = super().deferred_validate(data)
        data = verify_signature_upload(data)

        return data

    class Meta:
        model = CollectionVersionSignature
        fields = NoArtifactContentUploadSerializer.Meta.fields + (
            "signed_collection",
            "pubkey_fingerprint",
            "signing_service",
        )


class NamespaceLinkSerializer(serializers.Serializer):
    """
    Provides backwards compatible interface for links with the legacy
    GalaxyNG API.
    """

    url = serializers.URLField(max_length=256, allow_blank=False)
    name = serializers.CharField(max_length=256, allow_blank=False)


@extend_schema_field(NamespaceLinkSerializer(many=True))
class NamespaceLinkField(serializers.HStoreField):
    """
    Provides backwards compatible interface for links with the legacy
    GalaxyNG API.
    """

    def get_value(self, dictionary):
        data = dictionary.get(self.field_name, [])

        # The open api client sends data as a form request rather than json
        # because of the avatar URL. It converts a list like
        # [{"foo": "bar"}, {"bar": "foo"}] into "{'foo': 'bar'}, {'bar': 'foo'}"
        # This is a best effort attempt to capture that data and convert it into
        # valid JSON and then transform it into a list that the API can understand
        if isinstance(data, str):
            try:
                data = f"[{data}]".replace("'", '"')
                return json.loads(data)
            except json.decoder.JSONDecodeError:
                raise ValidationError(detail={"links": "Must be valid JSON"})

        return super().get_value(dictionary)

    def to_internal_value(self, data):
        if isinstance(data, dict):
            transformed = data
        else:
            transformed = {x["name"]: x["url"] for x in data}
        return super().to_internal_value(transformed)

    def to_representation(self, value):
        return [{"name": x, "url": value[x]} for x in value]


class AnsibleNamespaceMetadataSerializer(NoArtifactContentSerializer):
    """
    A serializer for Namespaces.
    """

    name = serializers.RegexField(
        NAME_REGEXP,
        min_length=3,
        max_length=64,
        allow_blank=False,
        help_text=_("Required named, only accepts lowercase, numbers and underscores."),
    )
    company = serializers.CharField(
        max_length=64,
        allow_blank=True,
        required=False,
        help_text=_("Optional namespace company owner."),
    )
    email = serializers.CharField(
        max_length=256,
        allow_blank=True,
        required=False,
        help_text=_("Optional namespace contact email."),
    )
    description = serializers.CharField(
        max_length=256, allow_blank=True, required=False, help_text=_("Optional short description.")
    )
    resources = serializers.CharField(
        allow_blank=True, required=False, help_text=_("Optional resource page in markdown format.")
    )
    links = NamespaceLinkField(
        child=serializers.URLField(max_length=256),
        required=False,
        help_text=_("Labeled related links."),
    )
    avatar = serializers.ImageField(
        write_only=True, required=False, help_text=_("Optional avatar image for Namespace")
    )
    avatar_sha256 = serializers.CharField(
        max_length=64, read_only=True, help_text=_("SHA256 digest of avatar image if present.")
    )
    avatar_url = serializers.SerializerMethodField(
        help_text=_("Download link for avatar image if present.")
    )
    metadata_sha256 = serializers.CharField(read_only=True)

    def get_avatar_url(self, obj):
        """Return the avatar url"""
        if obj.avatar_sha256:
            return settings.ANSIBLE_API_HOSTNAME + get_url(obj) + "avatar/"
        else:
            return None

    def validate(self, data):
        if self.instance:
            if (name := data.get("name", None)) and name != self.instance.name:
                raise serializers.ValidationError(_("Name can not be changed in an update"))

        # Check that avatar_sha256 is set if avatar was present in upload.
        if "artifact" in self.context:
            avatar_artifact = Artifact.objects.get(pk=self.context["artifact"])
            if data.get("avatar_sha256") is None:
                data["avatar_sha256"] = avatar_artifact.sha256
            elif data["avatar_sha256"] != avatar_artifact.sha256:
                raise serializers.ValidationError(_("Avatar does not match expected digest."))

        return super().validate(data)

    @transaction.atomic
    def create(self, validated_data):
        """Create the Namespace and add it to the Repository if present."""
        namespace, created = AnsibleNamespace.objects.get_or_create(name=validated_data["name"])
        metadata = AnsibleNamespaceMetadata(namespace=namespace, **validated_data)
        metadata.calculate_metadata_sha256()
        content = AnsibleNamespaceMetadata.objects.filter(
            metadata_sha256=metadata.metadata_sha256
        ).first()
        if content:
            content.touch()
        else:
            metadata.save()
            content = metadata
            if metadata.avatar_sha256:
                ContentArtifact.objects.create(
                    artifact_id=self.context["artifact"],
                    content=content,
                    relative_path=f"{metadata.name}-avatar",
                )

        repository = self.context.pop("repository", None)
        if repository:
            repository = AnsibleRepository.objects.get(pk=repository)
            content_to_add = AnsibleNamespaceMetadata.objects.filter(pk=content.pk)

            with repository.new_version() as new_version:
                new_version.add_content(content_to_add)
        return content

    class Meta:
        model = AnsibleNamespaceMetadata
        fields = (
            "pulp_href",
            "name",
            "company",
            "email",
            "description",
            "resources",
            "links",
            "avatar",
            "avatar_sha256",
            "avatar_url",
            "metadata_sha256",
        )


class CollectionVersionMarkSerializer(NoArtifactContentSerializer):
    """
    A serializer for mark models.
    """

    marked_collection = DetailRelatedField(
        help_text=_("The content this mark is pointing to."),
        view_name_pattern=r"content(-.*/.*)-detail",
        queryset=CollectionVersion.objects.all(),
    )
    value = serializers.SlugField(
        help_text=_("The string value of this mark."),
        allow_null=False,
    )

    class Meta:
        model = CollectionVersionMark
        fields = ["pulp_created", "pulp_href", "marked_collection", "value"]


class AnsibleRepositoryMarkSerializer(serializers.Serializer):
    """
    A serializer for the mark action.
    """

    content_units = serializers.ListField(
        required=True,
        help_text=_(
            "List of collection version hrefs to mark, use * to mark all content in repository"
        ),
    )
    value = serializers.SlugField(
        required=True,
        help_text=_("The string value of this mark."),
    )

    def validate_content_units(self, value):
        """Make sure the list is correctly formatted."""
        if len(value) > 1 and "*" in value:
            raise serializers.ValidationError("Cannot supply content units and '*'.")
        return value


class AnsibleRepositorySignatureSerializer(serializers.Serializer):
    """
    A serializer for the signing action.
    """

    content_units = serializers.ListField(
        required=True,
        help_text=_(
            "List of collection version hrefs to sign, use * to sign all content in repository"
        ),
    )
    signing_service = RelatedField(
        required=True,
        view_name="signing-services-detail",
        queryset=SigningService.objects.all(),
        help_text=_("A signing service to use to sign the collections"),
    )

    def validate_content_units(self, value):
        """Make sure the list is correctly formatted."""
        if len(value) > 1 and "*" in value:
            raise serializers.ValidationError("Cannot supply content units and '*'.")
        return value


class CollectionImportListSerializer(serializers.ModelSerializer):
    """
    A serializer for a CollectionImport list view.
    """

    id = serializers.UUIDField(source="pk")
    state = serializers.CharField(source="task.state")
    created_at = serializers.DateTimeField(source="task.pulp_created")
    updated_at = serializers.DateTimeField(source="task.pulp_last_updated")
    started_at = serializers.DateTimeField(source="task.started_at")
    finished_at = serializers.DateTimeField(source="task.finished_at", required=False)

    class Meta:
        model = CollectionImport
        fields = ("id", "state", "created_at", "updated_at", "started_at", "finished_at")


class CollectionImportDetailSerializer(CollectionImportListSerializer):
    """
    A serializer for a CollectionImport detail view.
    """

    error = fields.JSONDictField(source="task.error", required=False)
    messages = fields.JSONDictField()

    class Meta(CollectionImportListSerializer.Meta):
        fields = CollectionImportListSerializer.Meta.fields + ("error", "messages")


class CollectionVersionCopyMoveSerializer(serializers.Serializer):
    """
    Copy or move collections from a source repository into one or more destinations.

    This will carry associated content like Signatures and Marks along.
    """

    collection_versions = DetailRelatedField(
        required=True,
        view_name=r"content-ansible/collection_versions-detail",
        queryset=CollectionVersion.objects.all(),
        help_text=_("A list of collection versions to move or copy."),
        many=True,
    )

    destination_repositories = DetailRelatedField(
        required=True,
        view_name="repositories-ansible/ansible-detail",
        queryset=AnsibleRepository.objects.all(),
        help_text=_("List of repository HREFs to put content in."),
        many=True,
    )

    signing_service = RelatedField(
        required=False,
        view_name="signing-services-detail",
        queryset=SigningService.objects.all(),
        help_text=_(
            "HREF for a signing service. This will be used to sign the collection "
            "before moving putting it in any new repositories."
        ),
    )


class CopySerializer(serializers.Serializer):
    """
    A serializer for Content Copy API.
    """

    config = fields.JSONDictField(
        help_text=_("A JSON document describing sources, destinations, and content to be copied"),
    )

    def validate(self, data):
        """
        Validate that the Serializer contains valid data.

        Set the AnsibleRepository based on the RepositoryVersion if only the latter is provided.
        Set the RepositoryVersion based on the AnsibleRepository if only the latter is provided.
        Convert the human-friendly names of the content types into what Pulp needs to query on.
        """
        super().validate(data)

        if hasattr(self, "initial_data"):
            validate_unknown_fields(self.initial_data, self.fields)

        if "config" in data:
            validator = Draft7Validator(COPY_CONFIG_SCHEMA)

            err = []
            for error in sorted(validator.iter_errors(data["config"]), key=str):
                err.append(error.message)
            if err:
                raise serializers.ValidationError(
                    _("Provided copy criteria is invalid:'{}'".format(err))
                )

        return data



# Generated by Django 3.2.17 on 2023-02-08 21:50

import django.contrib.postgres.fields.hstore
from django.db import migrations, models
import django.db.models.deletion
import django_lifecycle.mixins
import uuid


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0099_versions_field'),
        ('ansible', '0046_add_fulltext_search_fix'),
    ]

    operations = [
        migrations.CreateModel(
            name='AnsibleNamespace',
            fields=[
                ('pulp_id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('name', models.CharField(max_length=64, unique=True)),
            ],
            options={
                'abstract': False,
            },
            bases=(django_lifecycle.mixins.LifecycleModelMixin, models.Model),
        ),
        migrations.CreateModel(
            name='AnsibleNamespaceMetadata',
            fields=[
                ('content_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, related_name='ansible_ansiblenamespacemetadata', serialize=False, to='core.content')),
                ('name', models.CharField(max_length=64)),
                ('company', models.CharField(blank=True, default='', max_length=64)),
                ('email', models.CharField(blank=True, default='', max_length=256)),
                ('description', models.CharField(blank=True, default='', max_length=256)),
                ('resources', models.TextField(blank=True, default='')),
                ('links', django.contrib.postgres.fields.hstore.HStoreField(default=dict)),
                ('avatar_sha256', models.CharField(max_length=64, null=True)),
                ('metadata_sha256', models.CharField(db_index=True, max_length=64)),
                ('namespace', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='metadatas', to='ansible.ansiblenamespace')),
            ],
            options={
                'default_related_name': '%(app_label)s_%(model_name)s',
                'unique_together': {('metadata_sha256',)},
            },
            bases=('core.content',),
        ),
    ]



from django.db import migrations
from django.db import models
from django.db.models import UniqueConstraint, Q
import semantic_version as semver


def migrate_collection_versions(apps, schema_editor):
    Collection = apps.get_model("ansible", "Collection")
    collections = Collection.objects.only('pk').all()
    for collection in collections:
        versions = collection.versions.only('pk', 'version').all()
        highest = max(versions, key=lambda x: semver.Version(x.version))
        highest.is_highest = True
        highest.save()


class Migration(migrations.Migration):
    dependencies = [
        ('ansible', '0004_add_fulltext_search_indexes'),
    ]

    operations = [
        migrations.AddField(
            model_name='collectionversion',
            name='is_highest',
            field=models.BooleanField(default=False, editable=False),
        ),
        migrations.AddConstraint(
            model_name='collectionversion',
            constraint=UniqueConstraint(
                fields=('collection', 'is_highest'),
                name='unique_is_highest',
                condition=Q(is_highest=True)
            ),
        ),
        migrations.RunPython(code=migrate_collection_versions,
                             reverse_code=migrations.RunPython.noop)
    ]



from unittest import mock
from django.test import TestCase

from orionutils.generator import randstr

from pulp_ansible.app.models import AnsibleDistribution
from pulp_ansible.app.models import AnsibleRepository
from pulp_ansible.app.models import CollectionVersion

from pulp_ansible.app.tasks.collections import (
    _rebuild_collection_version_meta,
    rebuild_repository_collection_versions_metadata,
    _update_highest_version,
)

from .utils import build_cvs_from_specs


class TestCollectionReImport(TestCase):
    """Test Collection Re-Import."""

    def setUp(self):
        """Make all the test data."""

        specs = [
            ("foo", "bar", "1.0.0"),
            ("foo", "baz", "1.0.0"),
            ("zip", "drive", "1.0.0"),
        ]

        self.collection_versions = build_cvs_from_specs(specs)

        # create a repository
        self.repo = AnsibleRepository(name="foorepo")
        self.repo.save()

        # create the distro
        AnsibleDistribution.objects.create(
            name="foorepo", base_path="foorepo", repository=self.repo
        )

        # add the cvs to a repository version
        for cv in self.collection_versions:
            qs = CollectionVersion.objects.filter(pk=cv.pk)
            with self.repo.new_version() as new_version:
                new_version.add_content(qs)

    def test_reimport_single_cv_adds_docs(self):
        """Test that rebuild_collection adds docs to the collection."""
        this_cv = self.collection_versions[0]
        _rebuild_collection_version_meta(this_cv)
        cv = CollectionVersion.objects.get(pulp_id=this_cv.pk)
        assert cv.docs_blob["collection_readme"]["html"] == "<h1>title</h1>\n<p>collection docs</p>"

    @mock.patch("pulp_ansible.app.tasks.collections._rebuild_collection_version_meta")
    @mock.patch("pulp_ansible.app.tasks.collections.ProgressReport")
    def test_reimport_repository_rebuilds_all_collections(
        self, mock_progress_report, mock_rebuild_cv
    ):
        """Make sure each CV in the repo is rebuilt."""
        rebuild_repository_collection_versions_metadata(self.repo.latest_version().pk)

        # ensure the function was called once for each cv
        expected_cvs = self.collection_versions
        assert mock_rebuild_cv.call_count == len(self.collection_versions)

        # ensure the appropriate pks were used
        call_pks = sorted([str(x.args[0].pk) for x in mock_rebuild_cv.mock_calls])
        expected_pks = sorted([str(x.pk) for x in expected_cvs])
        assert call_pks == expected_pks

    @mock.patch("pulp_ansible.app.tasks.collections._rebuild_collection_version_meta")
    @mock.patch("pulp_ansible.app.tasks.collections.ProgressReport")
    def test_reimport_repository_rebuilds_namespace(self, mock_progress_report, mock_rebuild_cv):
        """Make sure only the namespace CVs in the repo are rebuilt."""
        rebuild_repository_collection_versions_metadata(
            self.repo.latest_version().pk, namespace="foo"
        )

        # ensure the function was called once for each namespace cv
        expected_cvs = [x for x in self.collection_versions if x.namespace == "foo"]
        assert mock_rebuild_cv.call_count == len(expected_cvs)

        # ensure the appropriate pks were used
        call_pks = sorted([str(x.args[0].pk) for x in mock_rebuild_cv.mock_calls])
        expected_pks = sorted([str(x.pk) for x in expected_cvs])
        assert call_pks == expected_pks

    @mock.patch("pulp_ansible.app.tasks.collections._rebuild_collection_version_meta")
    @mock.patch("pulp_ansible.app.tasks.collections.ProgressReport")
    def test_reimport_repository_rebuilds_name(self, mock_progress_report, mock_rebuild_cv):
        """Make sure only the named CVs in the repo are rebuilt."""
        rebuild_repository_collection_versions_metadata(self.repo.latest_version().pk, name="baz")

        # ensure the function was called once for each namespace cv
        expected_cvs = [x for x in self.collection_versions if x.name == "baz"]
        assert mock_rebuild_cv.call_count == len(expected_cvs)

        # ensure the appropriate pks were used
        call_pks = sorted([str(x.args[0].pk) for x in mock_rebuild_cv.mock_calls])
        expected_pks = sorted([str(x.pk) for x in expected_cvs])
        assert call_pks == expected_pks

    def test_reimport_repository_rebuilds_contents(self):
        """Make sure the contents field in the CVs are rebuilt."""

        cobject = self.repo.latest_version().content.first()
        assert cobject is not None, "the repo fixture no longer has content for some unknown reason"

        # set some invalid contents for each CV in the repo ...
        cv = cobject.cast()
        cv.contents = ["a", "b", "c"]
        cv.save()

        # call the rebuild
        _rebuild_collection_version_meta(cobject)

        # after rebuild, the contents should have been changed back
        cv2 = cobject.cast()
        cv2.refresh_from_db()
        assert cv2.contents != ["a", "b", "c"]


class TestCollectionVersionHighest(TestCase):
    """Test Collection Re-Import."""

    def test_is_highest_set(self):
        namespace = randstr()
        name = randstr()
        specs = [
            (namespace, name, "1.0.1"),
        ]

        collection_versions = build_cvs_from_specs(specs, build_artifacts=False)
        for cv in collection_versions:
            _update_highest_version(cv)

        assert (
            CollectionVersion.objects.filter(namespace=namespace, name=name, version="1.0.1")
            .first()
            .is_highest
            is True
        )

    def test_is_highest_set_on_multple_stable_releases(self):
        namespace = randstr()
        name = randstr()
        specs = [
            (namespace, name, "1.0.1"),
            (namespace, name, "2.0.1"),
        ]

        collection_versions = build_cvs_from_specs(specs, build_artifacts=False)
        for cv in collection_versions:
            _update_highest_version(cv)

        assert (
            CollectionVersion.objects.filter(namespace=namespace, name=name, version="1.0.1")
            .first()
            .is_highest
            is False
        )

        assert (
            CollectionVersion.objects.filter(namespace=namespace, name=name, version="2.0.1")
            .first()
            .is_highest
            is True
        )

    def test_is_highest_set_on_single_prerelease(self):
        namespace = randstr()
        name = randstr()
        specs = [
            (namespace, name, "1.0.1-rc2"),
        ]

        collection_versions = build_cvs_from_specs(specs, build_artifacts=False)
        for cv in collection_versions:
            _update_highest_version(cv)

        assert (
            CollectionVersion.objects.filter(namespace=namespace, name=name, version="1.0.1-rc2")
            .first()
            .is_highest
            is True
        )

    def test_is_highest_set_on_multiple_prereleases(self):
        namespace = randstr()
        name = randstr()
        specs = [
            (namespace, name, "1.0.1-rc1"),
            (namespace, name, "1.0.1-rc2"),
            (namespace, name, "1.0.1-rc3"),
        ]

        collection_versions = build_cvs_from_specs(specs, build_artifacts=False)
        for cv in collection_versions:
            _update_highest_version(cv)

        assert (
            CollectionVersion.objects.filter(namespace=namespace, name=name, version="1.0.1-rc1")
            .first()
            .is_highest
            is False
        )

        assert (
            CollectionVersion.objects.filter(namespace=namespace, name=name, version="1.0.1-rc2")
            .first()
            .is_highest
            is False
        )

        assert (
            CollectionVersion.objects.filter(namespace=namespace, name=name, version="1.0.1-rc3")
            .first()
            .is_highest
            is True
        )

    def test_is_highest_set_on_multiple_prereleases_one_pass(self):
        namespace = randstr()
        name = randstr()
        specs = [
            (namespace, name, "1.0.1-rc1"),
            (namespace, name, "1.0.1-rc2"),
            (namespace, name, "1.0.1-rc3"),
        ]

        collection_versions = build_cvs_from_specs(specs, build_artifacts=False)
        _update_highest_version(collection_versions[1])

        assert (
            CollectionVersion.objects.filter(namespace=namespace, name=name, version="1.0.1-rc1")
            .first()
            .is_highest
            is False
        )

        assert (
            CollectionVersion.objects.filter(namespace=namespace, name=name, version="1.0.1-rc2")
            .first()
            .is_highest
            is False
        )

        assert (
            CollectionVersion.objects.filter(namespace=namespace, name=name, version="1.0.1-rc3")
            .first()
            .is_highest
            is True
        )

    def test_is_highest_set_on_multiple_prereleases_backwards_order(self):
        namespace = randstr()
        name = randstr()
        specs = [
            (namespace, name, "1.0.1-rc4"),
            (namespace, name, "1.0.1-rc1"),
            (namespace, name, "1.0.1-rc2"),
            (namespace, name, "1.0.1-rc3"),
        ]

        collection_versions = build_cvs_from_specs(specs, build_artifacts=False)
        for cv in collection_versions[::-1]:
            _update_highest_version(cv)

        assert (
            CollectionVersion.objects.filter(namespace=namespace, name=name, version="1.0.1-rc1")
            .first()
            .is_highest
            is False
        )

        assert (
            CollectionVersion.objects.filter(namespace=namespace, name=name, version="1.0.1-rc2")
            .first()
            .is_highest
            is False
        )

        assert (
            CollectionVersion.objects.filter(namespace=namespace, name=name, version="1.0.1-rc3")
            .first()
            .is_highest
            is False
        )

        assert (
            CollectionVersion.objects.filter(namespace=namespace, name=name, version="1.0.1-rc4")
            .first()
            .is_highest
            is True
        )

    def test_prerelease_not_higher_than_stable(self):
        namespace = randstr()
        name = randstr()
        specs = [
            (namespace, name, "1.0.1"),
            (namespace, name, "2.0.0-rc1"),
        ]

        collection_versions = build_cvs_from_specs(specs, build_artifacts=False)
        for cv in collection_versions:
            _update_highest_version(cv)

        assert (
            CollectionVersion.objects.filter(namespace=namespace, name=name, version="1.0.1")
            .first()
            .is_highest
            is True
        )

        assert (
            CollectionVersion.objects.filter(namespace=namespace, name=name, version="2.0.0-rc1")
            .first()
            .is_highest
            is False
        )

    def test_prerelease_matches_stable(self):
        namespace = randstr()
        name = randstr()
        specs = [
            (namespace, name, "2.0.0-rc1"),
            (namespace, name, "2.0.0"),
        ]

        collection_versions = build_cvs_from_specs(specs, build_artifacts=False)
        for cv in collection_versions:
            _update_highest_version(cv)

        assert (
            CollectionVersion.objects.filter(namespace=namespace, name=name, version="2.0.0-rc1")
            .first()
            .is_highest
            is False
        )

        assert (
            CollectionVersion.objects.filter(namespace=namespace, name=name, version="2.0.0")
            .first()
            .is_highest
            is True
        )



# Generated by Django 3.2.13 on 2022-07-01 10:19

import gnupg
import tempfile

from django.db import migrations, models


def migrate_keyring_to_gpgkey_on_repository_up(apps, schema_editor):
    AnsibleRepository = apps.get_model('ansible', 'AnsibleRepository')
    changed_repos = []
    with tempfile.TemporaryDirectory() as gnupghome:
        for repository in AnsibleRepository.objects.filter(keyring__isnull=False):
            try:
                gpg = gnupg.GPG(gnupghome=gnupghome, keyring=repository.keyring)
                repository.gpgkey = gpg.export_keys("*")
                changed_repos.append(repository)
            except Exception as e:
                # Make the migration resilient
                print(f"Repository {repository.pk} failed to migrate it's keyring. {e}")
                pass

            if len(changed_repos) >= 1024:
                AnsibleRepository.objects.bulk_update(changed_repos, ['gpgkey'])
                changed_repos.clear()
    # Flush the rest
    AnsibleRepository.objects.bulk_update(changed_repos, ['gpgkey'])


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0041_alter_collectionversion_collection'),
    ]

    operations = [
        migrations.AddField(
            model_name='ansiblerepository',
            name='gpgkey',
            field=models.TextField(null=True),
        ),
        migrations.RunPython(
            code=migrate_keyring_to_gpgkey_on_repository_up,
            reverse_code=migrations.RunPython.noop,
            elidable=True,
        ),
        migrations.RemoveField(
            model_name='ansiblerepository',
            name='keyring',
        ),
    ]



import pytest

from django.db import connection
from django.db.migrations.executor import MigrationExecutor


@pytest.fixture
def migrate(django_db_serialized_rollback):
    executor = MigrationExecutor(connection)

    def _migrate(target):
        connection.cursor().execute("SET CONSTRAINTS ALL IMMEDIATE;")
        state = executor.migrate(target)
        executor.loader.build_graph()
        return state.apps

    return _migrate



# Generated by Django 3.2.12 on 2022-04-01 13:59

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('ansible', '0040_ansiblerepository_keyring'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collectionversion',
            name='collection',
            field=models.ForeignKey(editable=False, on_delete=django.db.models.deletion.PROTECT, related_name='versions', to='ansible.collection'),
        ),
    ]



import pytest
from pulp_ansible.tests.functional.utils import iterate_all


@pytest.fixture
def repo_with_kitted_out_collection(
    pulpcore_bindings,
    ansible_bindings,
    build_and_upload_collection,
    ascii_armored_detached_signing_service,
    ansible_repo,
    ansible_distribution_factory,
    monitor_task,
):
    """
    Create a repo that has a collection with all the possible associated content.

    Repo contains 3 collections. 2 are returned to be copied, one of which has all
    the content associated with it and the other of which as a mark and signature.

    Returns the repo, two collections to copy and the list of expected content
    after the copy operation happens.
    """
    collection, collection_url = build_and_upload_collection()
    _, collection_url2 = build_and_upload_collection()
    _, non_copied_collection = build_and_upload_collection()

    # add collection to repo
    body = {"add_content_units": [collection_url, collection_url2, non_copied_collection]}
    monitor_task(ansible_bindings.RepositoriesAnsibleApi.modify(ansible_repo.pulp_href, body).task)

    # sign the collection
    body = {
        "content_units": [collection_url, collection_url2, non_copied_collection],
        "signing_service": ascii_armored_detached_signing_service.pulp_href,
    }
    monitor_task(ansible_bindings.RepositoriesAnsibleApi.sign(ansible_repo.pulp_href, body).task)

    # Mark the collection
    body = {
        "content_units": [collection_url, collection_url2, non_copied_collection],
        "value": "testable",
    }
    monitor_task(ansible_bindings.RepositoriesAnsibleApi.mark(ansible_repo.pulp_href, body).task)

    # Not gonna bother adding these to the first collection
    # Create a distro for the repo
    distro = ansible_distribution_factory(repository=ansible_repo)

    # Add a namespace for the collection
    kwargs = {"path": distro.base_path, "distro_base_path": distro.base_path}
    task = ansible_bindings.PulpAnsibleApiV3PluginAnsibleContentNamespacesApi.create(
        name=collection.namespace, description="hello", company="Testing Co.", **kwargs
    )
    monitor_task(task.task)

    # Deprecate the collection
    task = ansible_bindings.PulpAnsibleDefaultApiV3PluginAnsibleContentCollectionsIndexApi.update(
        name=collection.name,
        namespace=collection.namespace,
        distro_base_path=distro.base_path,
        patched_collection={"deprecated": True},
    )
    monitor_task(task.task)

    repo = ansible_bindings.RepositoriesAnsibleApi.read(ansible_repo.pulp_href)

    expected_content = {
        "collection_versions",
        "namespaces",
        "collection_signatures",
        "collection_marks",
        "collection_deprecations",
    }

    created_content = set()
    all_content = set()
    excluded_content = {non_copied_collection}

    # get the signatures and marks from the excluded collection:
    excluded_content = excluded_content.union(
        {
            x.pulp_href
            for x in iterate_all(
                ansible_bindings.ContentCollectionSignaturesApi.list,
                signed_collection=non_copied_collection,
            )
        }
    )

    excluded_content = excluded_content.union(
        {
            x.pulp_href
            for x in iterate_all(
                ansible_bindings.ContentCollectionMarksApi.list,
                marked_collection=non_copied_collection,
            )
        }
    )

    assert len(excluded_content) >= 3

    for content in iterate_all(
        pulpcore_bindings.ContentApi.list, repository_version=repo.latest_version_href
    ):
        created_content.add(content.pulp_href.strip("/").split("/")[-2])
        all_content.add(content.pulp_href)

    assert expected_content == created_content

    all_content = all_content.difference(excluded_content)

    return repo, [collection_url, collection_url2], all_content, excluded_content


@pytest.fixture
def repo_with_one_out_collection(
    ansible_bindings, build_and_upload_collection, ansible_repo, monitor_task
):
    """Create a distro serving two collections, one signed, one unsigned."""
    collection, collection_url = build_and_upload_collection()

    # add collection to repo
    body = {"add_content_units": [collection_url]}
    monitor_task(ansible_bindings.RepositoriesAnsibleApi.modify(ansible_repo.pulp_href, body).task)

    repo = ansible_bindings.RepositoriesAnsibleApi.read(ansible_repo.pulp_href)

    return repo, collection_url


def test_copy_with_all_content(
    pulpcore_bindings,
    ansible_bindings,
    repo_with_kitted_out_collection,
    ansible_repo_factory,
    monitor_task,
):
    """
    Test copying a collection with all it's associated content
    """
    src_repo, collections, all_content, excluded = repo_with_kitted_out_collection

    dest1 = ansible_repo_factory()
    dest2 = ansible_repo_factory()
    dest3 = ansible_repo_factory()

    dest = (dest1, dest2, dest3)

    task = ansible_bindings.RepositoriesAnsibleApi.copy_collection_version(
        src_repo.pulp_href,
        {
            "collection_versions": collections,
            "destination_repositories": [x.pulp_href for x in dest],
        },
    )

    monitor_task(task.task)

    for repo in dest:
        href = ansible_bindings.RepositoriesAnsibleApi.read(repo.pulp_href).latest_version_href
        dest_content = {
            x.pulp_href
            for x in iterate_all(pulpcore_bindings.ContentApi.list, repository_version=href)
        }

        assert dest_content == all_content

    href = ansible_bindings.RepositoriesAnsibleApi.read(src_repo.pulp_href).latest_version_href
    remaining = {
        c.pulp_href for c in iterate_all(pulpcore_bindings.ContentApi.list, repository_version=href)
    }

    assert remaining == excluded.union(all_content)


def test_move_with_all_content(
    pulpcore_bindings,
    ansible_bindings,
    repo_with_kitted_out_collection,
    ansible_repo_factory,
    monitor_task,
):
    """
    Test moving a collection with all it's associated content
    """
    src_repo, collections, all_content, excluded = repo_with_kitted_out_collection

    dest1 = ansible_repo_factory()
    dest2 = ansible_repo_factory()
    dest3 = ansible_repo_factory()

    dest = (dest1, dest2, dest3)

    task = ansible_bindings.RepositoriesAnsibleApi.move_collection_version(
        src_repo.pulp_href,
        {
            "collection_versions": collections,
            "destination_repositories": [x.pulp_href for x in dest],
        },
    )

    monitor_task(task.task)

    for repo in dest:
        href = ansible_bindings.RepositoriesAnsibleApi.read(repo.pulp_href).latest_version_href
        dest_content = {
            x.pulp_href
            for x in iterate_all(pulpcore_bindings.ContentApi.list, repository_version=href)
        }

        assert dest_content == all_content

    href = ansible_bindings.RepositoriesAnsibleApi.read(src_repo.pulp_href).latest_version_href
    remaining = {
        c.pulp_href for c in iterate_all(pulpcore_bindings.ContentApi.list, repository_version=href)
    }

    assert remaining == excluded


def test_copy_and_sign(
    pulpcore_bindings,
    ansible_bindings,
    ascii_armored_detached_signing_service,
    ansible_repo_factory,
    repo_with_one_out_collection,
    monitor_task,
):
    """Verify that you can copy and sign a collection."""
    src_repo, collection_url = repo_with_one_out_collection

    dest1 = ansible_repo_factory()
    dest2 = ansible_repo_factory()
    dest3 = ansible_repo_factory()

    dest = (dest1, dest2, dest3)

    task = ansible_bindings.RepositoriesAnsibleApi.copy_collection_version(
        src_repo.pulp_href,
        {
            "collection_versions": [collection_url],
            "destination_repositories": [x.pulp_href for x in dest],
            "signing_service": ascii_armored_detached_signing_service.pulp_href,
        },
    )

    monitor_task(task.task)

    for repo in dest:
        href = ansible_bindings.RepositoriesAnsibleApi.read(repo.pulp_href).latest_version_href
        assert pulpcore_bindings.ContentApi.list(repository_version=href).count == 2

        collections = ansible_bindings.ContentCollectionVersionsApi.list(repository_version=href)
        assert collections.count == 1

        signatures = ansible_bindings.ContentCollectionSignaturesApi.list(repository_version=href)
        assert signatures.count == 1

        assert signatures.results[0].signed_collection == collections.results[0].pulp_href


def test_move_and_sign(
    pulpcore_bindings,
    ansible_bindings,
    ascii_armored_detached_signing_service,
    ansible_repo_factory,
    repo_with_one_out_collection,
    monitor_task,
):
    """Verify that you can move and sign a collection."""
    src_repo, collection_url = repo_with_one_out_collection

    dest1 = ansible_repo_factory()
    dest2 = ansible_repo_factory()
    dest3 = ansible_repo_factory()

    dest = (dest1, dest2, dest3)

    task = ansible_bindings.RepositoriesAnsibleApi.move_collection_version(
        src_repo.pulp_href,
        {
            "collection_versions": [collection_url],
            "destination_repositories": [x.pulp_href for x in dest],
            "signing_service": ascii_armored_detached_signing_service.pulp_href,
        },
    )

    monitor_task(task.task)

    for repo in dest:
        href = ansible_bindings.RepositoriesAnsibleApi.read(repo.pulp_href).latest_version_href
        assert pulpcore_bindings.ContentApi.list(repository_version=href).count == 2

        collections = ansible_bindings.ContentCollectionVersionsApi.list(repository_version=href)
        assert collections.count == 1

        signatures = ansible_bindings.ContentCollectionSignaturesApi.list(repository_version=href)
        assert signatures.count == 1

        assert signatures.results[0].signed_collection == collections.results[0].pulp_href

    href = ansible_bindings.RepositoriesAnsibleApi.read(src_repo.pulp_href).latest_version_href
    assert pulpcore_bindings.ContentApi.list(repository_version=href).count == 0



# Generated by Django 4.2.1 on 2023-05-15 14:40

from django.db import migrations, models
import django.db.models.deletion
import pulpcore.app.models.base


class Migration(migrations.Migration):
    dependencies = [
        ("core", "0106_alter_artifactdistribution_distribution_ptr_and_more"),
        ("ansible", "0051_cvindex_build"),
    ]

    operations = [
        migrations.AlterField(
            model_name="ansiblecollectiondeprecated",
            name="content_ptr",
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to="core.content",
            ),
        ),
        migrations.AlterField(
            model_name="ansibledistribution",
            name="distribution_ptr",
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to="core.distribution",
            ),
        ),
        migrations.AlterField(
            model_name="ansiblenamespace",
            name="pulp_id",
            field=models.UUIDField(
                default=pulpcore.app.models.base.pulp_uuid,
                editable=False,
                primary_key=True,
                serialize=False,
            ),
        ),
        migrations.AlterField(
            model_name="ansiblenamespacemetadata",
            name="content_ptr",
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to="core.content",
            ),
        ),
        migrations.AlterField(
            model_name="ansiblerepository",
            name="repository_ptr",
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to="core.repository",
            ),
        ),
        migrations.AlterField(
            model_name="collection",
            name="pulp_id",
            field=models.UUIDField(
                default=pulpcore.app.models.base.pulp_uuid,
                editable=False,
                primary_key=True,
                serialize=False,
            ),
        ),
        migrations.AlterField(
            model_name="collectionremote",
            name="remote_ptr",
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to="core.remote",
            ),
        ),
        migrations.AlterField(
            model_name="collectionversion",
            name="content_ptr",
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to="core.content",
            ),
        ),
        migrations.AlterField(
            model_name="collectionversion",
            name="tags",
            field=models.ManyToManyField(editable=False, to="ansible.tag"),
        ),
        migrations.AlterField(
            model_name="collectionversionmark",
            name="content_ptr",
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to="core.content",
            ),
        ),
        migrations.AlterField(
            model_name="collectionversionsignature",
            name="content_ptr",
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to="core.content",
            ),
        ),
        migrations.AlterField(
            model_name="crossrepositorycollectionversionindex",
            name="collection_version",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="ansible.collectionversion"
            ),
        ),
        migrations.AlterField(
            model_name="crossrepositorycollectionversionindex",
            name="namespace_metadata",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="ansible.ansiblenamespacemetadata",
            ),
        ),
        migrations.AlterField(
            model_name="crossrepositorycollectionversionindex",
            name="repository",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="ansible.ansiblerepository"
            ),
        ),
        migrations.AlterField(
            model_name="crossrepositorycollectionversionindex",
            name="repository_version",
            field=models.ForeignKey(
                null=True, on_delete=django.db.models.deletion.CASCADE, to="core.repositoryversion"
            ),
        ),
        migrations.AlterField(
            model_name="downloadlog",
            name="pulp_id",
            field=models.UUIDField(
                default=pulpcore.app.models.base.pulp_uuid,
                editable=False,
                primary_key=True,
                serialize=False,
            ),
        ),
        migrations.AlterField(
            model_name="gitremote",
            name="remote_ptr",
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to="core.remote",
            ),
        ),
        migrations.AlterField(
            model_name="role",
            name="content_ptr",
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to="core.content",
            ),
        ),
        migrations.AlterField(
            model_name="roleremote",
            name="remote_ptr",
            field=models.OneToOneField(
                auto_created=True,
                on_delete=django.db.models.deletion.CASCADE,
                parent_link=True,
                primary_key=True,
                serialize=False,
                to="core.remote",
            ),
        ),
        migrations.AlterField(
            model_name="tag",
            name="pulp_id",
            field=models.UUIDField(
                default=pulpcore.app.models.base.pulp_uuid,
                editable=False,
                primary_key=True,
                serialize=False,
            ),
        ),
    ]



import pytest
from pulp_ansible.tests.functional.utils import content_counts
from pulpcore.client.pulp_ansible import AnsibleRepositorySyncURL


def test_add_mark_to_collection_version(
    ansible_bindings,
    ansible_repo_factory,
    build_and_upload_collection,
    monitor_task,
):
    """Test adding a mark to a collection version.

    1. Build and upload a collection
    2. Add it to a repo version
    3. Add a mark to the collection version
    4. Assert the mark was added correctly
    5. Remove the mark
    6. Assert the mark was removed correctly
    """
    repository = ansible_repo_factory()
    collection, collection_url = build_and_upload_collection()

    # add it to a repo version
    body = {"add_content_units": [collection_url]}
    monitor_task(ansible_bindings.RepositoriesAnsibleApi.modify(repository.pulp_href, body).task)
    repository = ansible_bindings.RepositoriesAnsibleApi.read(repository.pulp_href)

    # assert the content was added correctly
    assert repository.latest_version_href.endswith("/1/")

    repository_version = ansible_bindings.RepositoriesAnsibleVersionsApi.read(
        repository.latest_version_href
    )

    assert content_counts(repository_version) == {"ansible.collection_version": 1}

    # Add a mark to the collection version
    body = {"content_units": [collection_url], "value": "testable"}
    monitor_task(ansible_bindings.RepositoriesAnsibleApi.mark(repository.pulp_href, body).task)
    repository = ansible_bindings.RepositoriesAnsibleApi.read(repository.pulp_href)

    # assert the repo version incremented
    assert repository.latest_version_href.endswith("/2/")
    # Ensure mark with value "testable" is present on the marks
    marks = ansible_bindings.ContentCollectionMarksApi.list(
        marked_collection=collection_url
    ).results
    assert len(marks) == 1
    assert marks[0].value == "testable"
    assert marks[0].marked_collection == collection_url

    # Ensure mark is added to the repo
    repository_version = ansible_bindings.RepositoriesAnsibleVersionsApi.read(
        repository.latest_version_href
    )
    assert content_counts(repository_version, "added") == {"ansible.collection_mark": 1}

    # Unmark the collection version
    body = {"content_units": [collection_url], "value": "testable"}
    monitor_task(ansible_bindings.RepositoriesAnsibleApi.unmark(repository.pulp_href, body).task)

    # Refresh the repo version
    repository = ansible_bindings.RepositoriesAnsibleApi.read(repository.pulp_href)
    # ensure the repo version incremented
    assert repository.latest_version_href.endswith("/3/")
    # Ensure mark is removed from repo contents
    repository_version = ansible_bindings.RepositoriesAnsibleVersionsApi.read(
        repository.latest_version_href
    )
    assert content_counts(repository_version, "removed") == {"ansible.collection_mark": 1}


@pytest.fixture
def distro_serving_one_marked_one_unmarked_collection(
    ansible_bindings,
    ansible_repo_factory,
    build_and_upload_collection,
    ansible_distribution_factory,
    monitor_task,
):
    """Create a distro serving two collections, one marked, one unmarked."""
    repository = ansible_repo_factory()
    collections = []
    for i in range(2):
        _, collection_url = build_and_upload_collection()
        collections.append(collection_url)

    body = {"add_content_units": collections}
    monitor_task(ansible_bindings.RepositoriesAnsibleApi.modify(repository.pulp_href, body).task)

    body = {"content_units": collections[:1], "value": "marked-on-sync-test"}
    monitor_task(ansible_bindings.RepositoriesAnsibleApi.mark(repository.pulp_href, body).task)

    return ansible_distribution_factory(repository)


def test_sync_marks(
    ansible_bindings,
    ansible_repo_factory,
    ansible_collection_remote_factory,
    distro_serving_one_marked_one_unmarked_collection,
    monitor_task,
):
    """Test that marks are also synced."""
    distro = distro_serving_one_marked_one_unmarked_collection
    repository = ansible_repo_factory()

    # Create Remote
    remote = ansible_collection_remote_factory(url=distro.client_url, include_pulp_auth=True)

    # Sync
    repository_sync_data = AnsibleRepositorySyncURL(remote=remote.pulp_href)
    sync_response = ansible_bindings.RepositoriesAnsibleApi.sync(
        repository.pulp_href, repository_sync_data
    )
    monitor_task(sync_response.task)
    repository = ansible_bindings.RepositoriesAnsibleApi.read(repository.pulp_href)

    repository_version = ansible_bindings.RepositoriesAnsibleVersionsApi.read(
        repository.latest_version_href
    )
    assert content_counts(repository_version) == {
        "ansible.collection_version": 2,
        "ansible.collection_mark": 1,
    }



"""Tests related to Galaxy V3 deprecation."""

import pytest


@pytest.mark.parallel
@pytest.mark.parametrize("repo_kwargs", [{}, {"retain_repo_versions": 1}])
def test_deprecation(
    ansible_bindings,
    ansible_collection_remote_factory,
    ansible_repo_factory,
    ansible_distribution_factory,
    ansible_sync_factory,
    monitor_task,
    repo_kwargs,
):
    """Test sync  sync."""
    # Sync down two collections into a repo
    requirements = (
        "collections:\n" "  - name: testing.k8s_demo_collection\n" "  - name: pulp.squeezer"
    )

    first_remote = ansible_collection_remote_factory(
        url="https://galaxy.ansible.com",
        requirements_file=requirements,
        sync_dependencies=False,
    )

    first_repo = ansible_sync_factory(
        ansible_repo_factory(remote=first_remote.pulp_href, **repo_kwargs)
    )
    first_distribution = ansible_distribution_factory(repository=first_repo)

    # Assert the state of deprecated True for testing, False for pulp
    collections = ansible_bindings.PulpAnsibleApiV3CollectionsApi.list(
        first_distribution.base_path, namespace="testing"
    )

    assert collections.data[0].deprecated
    collections = ansible_bindings.PulpAnsibleApiV3CollectionsApi.list(
        first_distribution.base_path, namespace="pulp"
    )
    assert not collections.data[0].deprecated

    # Sync a second repo from the first, just the testing namespace
    requirements = "collections:\n" "  - name: testing.k8s_demo_collection"
    second_remote = ansible_collection_remote_factory(
        url=first_distribution.client_url,
        requirements_file=requirements,
        sync_dependencies=False,
        include_pulp_auth=True,
    )

    second_repo = ansible_sync_factory(
        ansible_repo_factory(remote=second_remote.pulp_href, **repo_kwargs)
    )
    second_distribution = ansible_distribution_factory(repository=second_repo)

    # Ensure the second remote received a deprecated=True for the testing namespace collection
    collections = ansible_bindings.PulpAnsibleApiV3CollectionsApi.list(
        second_distribution.base_path, namespace="testing"
    )
    assert collections.data[0].deprecated

    # Change the deprecated status for the testing collection on the original repo to False
    result = ansible_bindings.PulpAnsibleApiV3CollectionsApi.update(
        "k8s_demo_collection", "testing", first_distribution.base_path, {"deprecated": False}
    )
    monitor_task(result.task)
    collections = ansible_bindings.PulpAnsibleApiV3CollectionsApi.list(
        first_distribution.base_path, namespace="testing"
    )
    assert not collections.data[0].deprecated

    # Update the requirements to sync down both collections this time
    requirements = (
        "collections:\n" "  - name: testing.k8s_demo_collection\n" "  - name: pulp.squeezer"
    )
    ansible_bindings.RemotesCollectionApi.partial_update(
        second_remote.pulp_href, {"requirements_file": requirements}
    )

    # Sync the second repo again
    second_repo = ansible_sync_factory(second_repo)

    # Assert both collections show deprecated=False
    collections = ansible_bindings.PulpAnsibleApiV3CollectionsApi.list(
        second_distribution.base_path
    )
    for collection in collections.data:
        assert not collection.deprecated



"""Tests related to sync ansible plugin collection content type."""

import pytest

from orionutils.generator import randstr

from pulp_ansible.tests.functional.constants import (
    ANSIBLE_DEMO_COLLECTION_REQUIREMENTS as DEMO_REQUIREMENTS,
    GALAXY_ANSIBLE_BASE_URL,
)


@pytest.mark.parallel
def test_tags_filter(
    ansible_bindings,
    ansible_collection_remote_factory,
    ansible_sync_factory,
):
    """Filter CollectionVersions by using the tags filter.

    This test targets the following issue:

    * `Pulp #5571 <https://pulp.plan.io/issues/5571>`_

    Do the following:

    1. Create a repository, and a remote.
    2. Sync the remote.
    3. Attempt to filter the CollectionVersions by different tags

    Note that the testing.k8s_demo_collection collection has tags 'k8s' and 'kubernetes'.
    """
    ansible_sync_factory(
        remote=ansible_collection_remote_factory(
            url=GALAXY_ANSIBLE_BASE_URL, requirements_file=DEMO_REQUIREMENTS
        ).pulp_href
    )

    # filter collection versions by tags
    params = {"tags": "nada"}
    collections = ansible_bindings.ContentCollectionVersionsApi.list(**params).results
    assert len(collections) == 0, collections

    params = {"tags": "k8s"}
    collections = ansible_bindings.ContentCollectionVersionsApi.list(**params).results
    assert len(collections) == 1, collections

    params = {"tags": "k8s,kubernetes"}
    collections = ansible_bindings.ContentCollectionVersionsApi.list(**params).results
    assert len(collections) == 1, collections

    params = {"tags": "nada,k8s"}
    collections = ansible_bindings.ContentCollectionVersionsApi.list(**params).results
    assert len(collections) == 0, collections


@pytest.mark.parallel
def test_content_unit_lifecycle(ansible_bindings, build_and_upload_collection, monitor_task):
    """Create content unit."""
    attrs = {"namespace": randstr(), "name": "squeezer", "version": "0.0.9"}
    collection_artifact, content_unit_href = build_and_upload_collection(config=attrs)

    # Read a content unit by its href.
    content_unit = ansible_bindings.ContentCollectionVersionsApi.read(content_unit_href)
    for key, val in attrs.items():
        assert content_unit.to_dict()[key] == val

    # Read a content unit by its pkg_id.
    page = ansible_bindings.ContentCollectionVersionsApi.list(
        namespace=content_unit.namespace, name=content_unit.name
    )
    assert page.count == 1
    assert page.results[0] == content_unit

    # Attempt to update a content unit using HTTP PATCH.
    # This HTTP method is not supported and a HTTP exception is expected.
    with pytest.raises(AttributeError) as exc_info:
        ansible_bindings.ContentCollectionVersionsApi.partial_update(
            content_unit_href, name="testing"
        )
    msg = "object has no attribute 'partial_update'"
    assert msg in exc_info.value.args[0]

    # Attempt to update a content unit using HTTP PUT.
    # This HTTP method is not supported and a HTTP exception is expected.
    with pytest.raises(AttributeError) as exc_info:
        ansible_bindings.ContentCollectionVersionsApi.update(content_unit_href, {"name": "testing"})
    msg = "object has no attribute 'update'"
    assert msg in exc_info.value.args[0]

    # Attempt to delete a content unit using HTTP DELETE.
    # This HTTP method is not supported and a HTTP exception is expected.
    with pytest.raises(AttributeError) as exc_info:
        ansible_bindings.ContentCollectionVersionsApi.delete(content_unit_href)
    msg = "object has no attribute 'delete'"
    assert msg in exc_info.value.args[0]

    # Attempt to create duplicate collection.
    create_task = monitor_task(
        ansible_bindings.ContentCollectionVersionsApi.create(file=collection_artifact.filename).task
    )
    assert content_unit_href in create_task.created_resources
